"""
logging.py log file making module
"""
from __future__ import absolute_import, division, print_function

import sys
import os
import time
import datetime
import copy
import io

from collections import deque
from collections.abc import MutableSequence, MutableMapping, Mapping

from ..aid.sixing import *
from .globaling import *
from ..aid.odicting import odict
from ..aid.filing import ocfn

from . import excepting
from . import registering
from . import storing
from . import tasking

from ..aid.consoling import getConsole
console = getConsole()

#Class definitions


class Logger(tasking.Tasker):
    """
    Logger Task Patron Registry Class for managing Logs

    Usage:   logger.send(START) to prepare log formats also reopens files
             logger.send(RUN) runs logs
             logger.send(STOP) closes log files needed to flush caches

    """

    def __init__(self,
                 flushPeriod=30.0,
                 prefix='~/.ioflo/log',
                 keep=0,
                 cyclePeriod=0.0,
                 fileSize=0,
                 reuse=False,
                 **kw):
        """
        Initialize instance.

        Inherited Parameters:
            name = unique name for logger
            store = data store
            period = time period between runs of logger
            schedule = tasker shedule such as ACTIVE INACTIVE

        Parameters:
            flushPeriod = time in seconds between flushes
            prefix = prefix used to create log directory
            keep = int number of log copies in rotation <1> means do not cycle
            cyclePeriod = interval in seconds between log rotations,
                     0.0 or None means do not rotate
            fileSize = size in bytes  of log file required to peform rotation
                       Do not rotate is main file is not at least meet file size
                       0 means always rotate
            reuse = Make unique time stamped log directory if True otherwise nonunique
                    useful when rotating


        Inherited Class Attributes:
            Counter = number of instances in class registrar
            Names = odict of instances keyed by name in class registrar

        Inherited instance attributes
            .name = unique name for logger
            .store = data store for house
            .period = desired time in seconds between runs,non negative, zero means asap
            .schedule = initial scheduling context for this logger vis a vis skedder

            .stamp = depends on subclass default is time logger last RUN
            .status = operational status of logger
            .desire = desired control asked by this or other taskers
            .done = logger completion state True or False
            .runner = generator to run logger

        Instance attributes
            .flushPeriod = period between flushes
            .prefix = prefix used to create log directory
            .keep = int number of log copies in rotation, < 1 means do cycle
            .cyclePeriod = interval in seconds between log rotations,
                     0.0 or None means do not rotate
            .fileSize = minimum size in bytes of main log file for rotation to occur
            .reuse = Make unique time stamped log directory if True otherwise nonunique
                    useful when rotating

            .rotateStamp = time logs last rotated
            .flushStamp = time logs last flushed
            .path = full path name of log directory
            .logs = dict of logs


        """
        super(Logger, self).__init__(**kw) #status = STOPPED  make runner advance so can send cmd

        self.flushPeriod = max(1.0, flushPeriod)
        self.prefix = prefix #prefix to log directory path
        self.keep = int(keep)
        self.cyclePeriod = max(0.0, cyclePeriod)  # ensure >= 0
        if self.keep > 0 and not self.cyclePeriod:
            self.keep = 0  # cyclePeriod must be nonzero if keep > 0
        self.fileSize = max(0, fileSize)
        self.reuse = True if reuse else False

        self.cycleStamp = 0.0
        self.flushStamp = 0.0
        self.path = '' #log directory path created on .reopen()
        self.logs = [] #list of logs

    def log(self):
        """
        Perform one log action
        """
        for log in self.logs:
            log()

        try:
            if (self.store.stamp - self.flushStamp) >= self.flushPeriod:
                console.profuse("Logger {0} Flushed at {1}, previous flush at {2}\n".format(
                    self.name, self.store.stamp, self.flushStamp))
                self.flush()
                self.flushStamp = self.store.stamp

        except TypeError:  # stamps may be None so handle
            self.flushStamp = self.store.stamp  # force flushStamp to be store.stamp

        if self.keep:
            try:
                if (self.store.stamp - self.cycleStamp) >= self.cyclePeriod:
                    console.profuse("Logger {0} Cycle rotation at {1}, previous cycle at {2}\n".format(
                        self.name, self.store.stamp, self.cycleStamp))
                    self.cycle()
                    self.cycleStamp = self.store.stamp

            except TypeError:  # stamps may be None so handle
                self.cycleStamp = self.store.stamp  # force cycleStamp to be store.stamp

    def reopen(self):
        """
        Reopen all log files
        """
        if not self.path:  # not created yet
            self.path = self.createPath(prefix = self.prefix)
            if not self.path:
                return False

        for log in self.logs:
            if not log.reopen(prefix=self.path, keep=self.keep):
                return False

        return True

    def close(self):
        """
        Close all log files
        """
        for log in self.logs:
            log.close()

    def flush(self):
        """
        Flush all log files
        """
        for log in self.logs:
            log.flush()

    def cycle(self):
        """
        Cycle (Rotate) all log files
        """
        for log in self.logs:
            log.cycle(size=self.fileSize)

    def prepare(self):
        """
        Called in runner on control = START
        """
        for log in self.logs:
            log.prepare()

    def resolve(self):
        """
        Called by house to resolve links in tasker
        """
        for log in self.logs:
            log.resolve()

    def addLog(self, log):
        """
        Add log to list of logs
        """
        self.logs.append(log)

    def createPath(self, prefix = '~/.ioflo/log'):
        """
        Returns unique logger base directory path
        if successfully creates base logger directory, empty otherwise
        creates unique log directory path
        creates physical directories on disk
        """
        path = ''
        try:
            if self.reuse:
                path = os.path.join(prefix, self.store.house.name, self.name)
                path = os.path.abspath(path)  # convert to proper absolute path
                if not os.path.exists(path):
                    os.makedirs(path)

            else:  # unique directory name
                i = 0
                while True:  # do until keep trying until different

                    dt = datetime.datetime.now()
                    dirname = "{0}_{1:04d}{2:02d}{3:02d}_{4:02d}{5:02d}{6:02d}_{7:03d}".format(
                               self.name, dt.year, dt.month, dt.day, dt.hour,
                               dt.minute, dt.second, dt.microsecond // 1000 + i )

                    path = os.path.join(prefix, self.store.house.name, dirname)
                    path = os.path.abspath(path) #convert to proper absolute path
                    if not os.path.exists(path):
                        os.makedirs(path)
                        break
                    i +=1

        except OSError as ex:
            console.terse("Error: creating log directory '{0}'\n".format(ex))
            return path

        console.concise("     Created Logger {0} Directory at '{1}'\n".format(
                self.name, self.path))

        return path

    def makeRunner(self):
        """
        generator factory function to create generator to run this logger
        """
        #do any on creation initialization here
        console.profuse("     Making Logger Task Runner {0}\n".format(self.name))

        self.status = STOPPED #operational status of tasker
        self.desire = STOP #default what to do next time, override below

        try: #catch exceptions to close log files before exiting generator
            while (True):
                control = (yield (self.status )) #accept control and yield status
                console.profuse("\n     Iterate Logger {0} with control = {1} status = {2}\n".format(
                    self.name,
                    ControlNames.get(control, 'Unknown'),
                    StatusNames.get(self.status, 'Unknown')))

                if control == RUN:
                    console.profuse("     Running Logger {0} ...\n".format(self.name))
                    self.log()
                    self.status = RUNNING

                elif control == READY:
                    console.profuse("     Attempting Ready Logger {0}\n".format(self.name))
                    #doesn't do anything yet
                    console.terse("     Readied Logger {0} ...\n".format(self.name))
                    self.status = READIED

                elif control == START:
                    console.profuse("     Attempting Start Logger {0}\n".format(self.name))

                    if self.reopen():
                        console.terse("     Starting Logger {0} ...\n".format(self.name))
                        self.prepare()
                        self.log()
                        self.desire = RUN
                        self.status = STARTED
                    else:
                        self.desire = STOP
                        self.status = STOPPED

                elif control == STOP:
                    if self.status != STOPPED:
                        console.terse("     Stopping Logger {0} ...\n".format(self.name))
                        self.log() #final log
                        if self.keep and self.reuse:  # recycle in case multiple restarts
                            self.cycle()  # cause log file to exceed size before cycle time
                        self.close()
                        self.desire = STOP
                        self.status = STOPPED

                else:  #control == ABORT
                    console.profuse("     Aborting Logger {0} ...\n".format(self.name))
                    self.close()
                    self.desire = ABORT
                    self.status = ABORTED

                self.stamp = self.store.stamp

        except Exception as ex:
            console.terse("{0}\n".format(ex))
            console.terse("     Exception in Logger {0} in {1}\n".format(
                    self.name, self.store.house.name))
            raise

        finally:
            self.close() #close all log files
            self.desire = ABORT
            self.status = ABORTED

class Log(registering.StoriedRegistrar):
    """
    Log Class for logging to file

    Iherited instance attributes:
       .name = unique name for log (group)
       .store = data store

    Instance attributes:
       .stamp = time stamp last time logged used by once and update actions
       .kind = text or binary
       .fileName = file name only
       .path = full dir path name of file
       .file = file where log is written
       .rule = log rule conditions for log
       .action = function to use when logging
       .header = header for log file

       .loggees = odict of loggee shares to be logged keyed by tag
       .formats = odict of loggee format string odicts keyed by tag
                  each format string odict is are format strings keyed by data field
       .lasts = odict of loggee Data instances of last values keyed by tag
                  each Data instance attribute is data field
    """
    Counter = 0  # Logs have their own namespace
    Names = {}

    def __init__(self,
                 kind='text',
                 baseFilename='',
                 rule=NEVER,
                 loggees=None,
                 fields=None,
                 **kw):
        """
        Initialize instance.
        Parameters:
            kind = text or binary
            baseFilename = base for log file name, extension added later when path created
            rule = log rule conditions (NEVER, ONCE, ALWAYS, UPDATE, CHANGE)
            loggees = odict of shares to be logged keyed by tags
            fields = odict of field name lists keyed by loggee tag

        """
        if 'preface' not in kw:
            kw['preface'] = 'Log'

        super(Log,self).__init__(**kw) #store and name inited here

        self.stamp = None  # time stamp last logged, None means not yet logged
        self.first = True  # True means file created for the first time

        self.kind = kind
        if baseFilename:
            self.baseFilename = baseFilename #file name only
        else:
            self.baseFilename = self.name
        self.path = ''  # full dir path name of file
        self.file = None  # file where log is written
        self.paths = []  # file path names of log rotate copies

        self.rule = rule #log rule when to log
        self.action = None #which method to use when logging
        self.assignRuleAction() #assign log action function

        self.header = ''  # log file header

        self.loggees = odict()  # odict of share refs to be logged (loggees) keyed by tag
        # odict of lists of field names for loggee keyed by tag
        self.fields =  fields if fields is not None else odict()
        self.formats = odict()  # odict of format string odicts keyed by tag
                                # each entry value is odict of format strings keyed by data field
        self.lasts = odict()  # odict of data instances of last values  keyed by tag

        if loggees:
            for tag, loggee in loggees.items():
                self.addLoggee(tag, loggee)


    def resolve(self):
        """
        resolves links to loggees

        """
        console.profuse("     Resolving links for Log {0}\n".format(self.name))

        for tag, loggee in self.loggees.items():
            if not isinstance(loggee, storing.Share):
                share = self.store.fetch(loggee)
                if share is None:
                    raise excepting.ResolveError("Loggee not in store", loggee, self.name)
                self.loggees[tag] = share #replace link name with link

    def __call__(self, **kw):
        """
        run .action
        """
        self.action(**kw)
        console.profuse("     Log {0} at {1}\n".format(self.name, self.stamp))

    def createPath(self, prefix):
        """
        Returns full path name of file given prefix and .kind .baseFilename
        """
        if self.kind == 'text':
            ext = '.txt'
        elif self.kind == 'binary':
            ext = '.log'

        filename= "{0}{1}".format(self.baseFilename, ext)
        path = os.path.join(prefix, filename)
        path = os.path.abspath(path)  # convert to proper absolute path
        return path

    def reopen(self, prefix='', keep=0):
        """
        Returns True is successful False otherwise
        Closes if open then reopens
        Opens or Creates log file and assign to .path
        If .path empty then creates path using prefix
        If keep then creates cycle (rotation) copy paths in .paths
           trial opens cycle paths.
        """
        keep = int(keep)

        if not self.path:
            self.path = self.createPath(prefix=prefix)
            self.paths = []  # remove stale rotate paths

        self.close()  #innocuous to call close() on unopened file
        if os.path.exists(self.path):
            self.first = False

        try:
            self.file = ocfn(self.path, 'a+')  # append pick up where left off
        except IOError as ex:
            console.terse("Error: Creating/opening log file '{0}'\n".format(ex))
            self.file = None
            return False

        console.concise("     Created/Opened Log file '{0}'\n".format(self.path))

        if keep > 0:
            self.paths = [self.path]
            for k in range(keep):
                k += 1
                root, ext = os.path.splitext(self.path)
                path = "{0}{1:02}{2}".format(root, k, ext)
                self.paths.append(path)

                try:  # trial open file to make
                    file = ocfn(path, 'r')  # do not truncate in case reusing
                except IOError as ex:
                    console.terse("Error: Creating/opening log rotate file '{0}'\n".format(ex))
                    return False
                file.close()
                console.concise("     Created Log rotate file '{0}'\n".format(path))

        return True

    def close(self):
        """
        close self.file if open except stdout
        """
        if self.file and not self.file.closed:
            self.flush()  # close does not necessarily fsync
            self.file.close()
            self.file = None

    def flush(self):
        """
        flush self.file if open except stdout
        """
        if self.file and not self.file.closed:
            self.file.flush()
            os.fsync(self.file.fileno())

    def cycle(self, size=0):
        """
        Returns True if cycle rotate successful, False otherwise
        Cycle log files  Only cycle if size > 0 and main log file size >= size
        """

        if self.paths:  # non zero rotate copies
            self.flush()
            try:
                if size and os.path.getsize(self.path) < size:
                    return False
            except OSError as ex:
                console.terse("Error: Reading file size '{0}'\n".format(ex))
                return False

            self.close()  # also flushes
            cycled = True
            for k in reversed(range(len(self.paths) - 1)):
                old = self.paths[k]
                new = self.paths[k+1]
                try:
                    os.rename(old, new)
                except OSError as ex:
                    console.terse("Error: Moving log rotate file '{0}'\n".format(ex))
                    cycled = False
                    break

            if not cycled:
                self.reopen()  # reopen so don't lose data
                return False

            try:  # truncate main file
                self.file = ocfn(self.path, 'w+')  # truncate.file
            except IOError as ex:
                console.terse("Error: Truncating log file '{0}'\n".format(ex))
                self.file = None
                return False

            self.file.write(self.header)  # rewrite header
            self.reopen()  # reopen for append

        return True

    def assignRuleAction(self, rule = None):
        """
        Assigns correct log action based on rule
        """
        #should be different if binary kind
        if rule is not None:
            self.rule = rule

        if self.rule == ONCE:
            self.action = self.once
        elif self.rule == ALWAYS:
            self.action = self.always
        elif self.rule == UPDATE:
            self.action = self.update
        elif self.rule == CHANGE:
            self.action = self.change
        elif self.rule == STREAK:
            self.action = self.streak
        elif self.rule == DECK:
            self.action = self.deck
        else:
            self.action = self.never

    def buildHeader(self):
        """
        Build  .header
        """
        # build first line header with kind rule and file name
        cf = io.StringIO()
        cf.write(ns2u(self.kind))
        cf.write(u'\t')
        cf.write(ns2u(LogRuleNames[self.rule]))
        cf.write(u'\t')
        cf.write(ns2u(self.baseFilename))
        cf.write(u'\n')
        # build second line header with field names
        cf.write(u'_time')
        for tag, fields in self.fields.items():
            if len(fields) > 1:  # multiple data fields so prepend with tag.
                for field in fields:
                    cf.write(u"\t{0}.{1}".format(tag, field))
            else:
                cf.write(u"\t{0}".format(tag))

        cf.write(u'\n')
        self.header = cf.getvalue()
        cf.close()

    def prepare(self):
        """
        Prepare log formats and values
        """
        console.profuse("     Preparing formats for Log {0}\n".format(self.name))

        if self.rule in (DECK, ):
            tag, loggee = self.loggees.items()[0]  # first loggee only
            fields = self.fields.get(tag)
            if not fields:
                raise ValueError("Log {0}: Rule '{1}' requires field list.".format(
                    self.name,
                    LogRuleNames[DECK]))

        if self.rule in (STREAK, ):
            tag, loggee = self.loggees.items()[0]  # first loggee only
            if tag in self.fields:
                if self.fields[tag]:  # at least one field
                    self.fields[tag] = self.fields[tag][:1]  # only one field allowed
                else:  # use first fields in loggee if present
                    self.fields[tag] = [loggee.keys()[0]] if loggee else []
            else:  # fields not given use first fields in loggee if present
                self.fields[tag] = [loggee.keys()[0]] if loggee else []


        else:  # default fields from loggee
            for tag, loggee in self.loggees.items():
                if tag not in self.fields or not self.fields[tag]:  # if fields not given use all fields in loggee
                    self.fields[tag] = [field for field in loggee]

        #should be different if binary kind
        #build formats
        self.formats.clear()
        self.formats['_time'] = '%s'  # '%0.6f'
        for tag, fields in self.fields.items():
            self.formats[tag] = odict()
            for field in fields:
                fmt = '\t%s'  # str for all
                self.formats[tag][field] = fmt

        #if self.rule in (STREAK, DECK, ):  # no way to know data so default fmt
            #for tag, fields in self.fields.items():
                #self.formats[tag] = odict()
                #for field in fields:
                    #fmt = '\t%s'
                    #self.formats[tag][field] = fmt

        #else:  # formats from loggee fields if present
            #for tag, fields in self.fields.items():
                #self.formats[tag] = odict()
                #loggee = self.loggees[tag]
                #for field in fields:
                    #if field in loggee:
                        #fmt = self.format(loggee[field])
                    #else:
                        #fmt = '\t%s'
                    #self.formats[tag][field] = fmt


        if self.rule in (CHANGE, ):  # build last copies for if changed
            self.lasts.clear()
            for tag, fields in self.fields.items():  # list of fields by tag
                loggee = self.loggees[tag]
                lasts = [(key, loggee[key]) for key in fields if key in loggee]
                self.lasts[tag] = storing.Data(lasts)  # in both loggee and fields

        self.buildHeader()

        if self.stamp is None and self.first:  # never logged so log headers
            self.file.write(self.header)

    def format(self, value):
        """
        returns format string for value type
        """
        if isinstance(value, float):
            return '\t%0.4f'
        elif isinstance(value, bool):
            return '\t%s'
        elif isinstance(value, int):
            return '\t%d'
        else:
            return '\t%s'

    def log(self):
        """
        log loggees
        called by conditional actions
        """
        self.stamp = self.store.stamp

        #should be different if binary kind
        cf = io.StringIO() #use string io faster than concatenation
        try:
            text = self.formats['_time'] % self.stamp
        except TypeError:
            text = '%s' % self.stamp
        cf.write(ns2u(text))

        for tag, loggee in self.loggees.items():
            for field, fmt in self.formats[tag].items():
                if field in loggee:
                    value = loggee[field]
                    try:
                        text = fmt % value
                    except TypeError:
                        text = '\t%s' % value
                    cf.write(ns2u(text))

                else:  # field no longer present in loggee so just tab
                    cf.write(u'\t')

        cf.write(u'\n')

        try:
            self.file.write(cf.getvalue())
        except ValueError as ex: #if self.file already closed then ValueError
            console.terse("{0}\n".format(ex))

        cf.close()

    def logStreak(self):
        """
        called by conditional actions
        Log and remove all elements of sequence in fifo order
        head is left tail is right, Fifo is head to tail

        """
        self.stamp = self.store.stamp

        #should be different if binary kind
        cf = io.StringIO() #use string io faster than concatenation

        if self.loggees:
            tag, loggee = self.loggees.items()[0] # only works for one loggee
            if loggee: # not empty has at least one field
                if not self.fields[tag]:  # field was not prepared
                    field = loggee.keys()[0]  # first field
                    fmt = "\t%s"  # default
                else:
                    field = self.fields[tag][0]  # first prepared field
                    fmt = self.formats[tag][field]

                if field in loggee:
                    value = loggee[field]
                    d = deque()
                    if isinstance(value, MutableSequence): # has pop method
                        while value: # not empty
                            d.appendleft(value.pop()) #remove and copy in order

                    elif isinstance(value, MutableMapping): # has popitem method
                        while value: # not empty
                            d.appendleft(value.popitem()) #remove and copy in order

                    else: #not mutable sequence or mapping so log normally
                        d.appendleft(value)

                    while d: # not empty
                        try:
                            text = self.formats['_time'] % self.stamp
                        except TypeError:
                            text = '%s' % self.stamp
                        cf.write(ns2u(text))

                        element = d.popleft()
                        try:
                            text = fmt % (element, )
                        except TypeError:
                            text = '\t%s' % element
                        cf.write(ns2u(text))
                        cf.write(u'\n')

                    try:
                        self.file.write(cf.getvalue())
                    except ValueError as ex: #if self.file already closed then ValueError
                        console.terse("{0}\n".format(ex))

        cf.close()

    def logDeck(self):
        """
        called by conditional actions
        Log and remove all elements of deck in fifo order which is pull
        """
        self.stamp = self.store.stamp

        if self.loggees:
            tag, loggee = self.loggees.items()[0]  # only works for first loggee
            fields = self.fields[tag]

            if loggee.deck:  # something to log
                cf = io.StringIO() #use string io faster than concatenation

                while loggee.deck:  # while not empty deck
                    entry = loggee.pull()  # assumed a dict
                    if not isinstance(entry, Mapping):
                        console.concise("Log {0}: Deck entry of '{1}' = '{2}' not a "
                                    "mapping.\n".format(self.name, loggee.name, entry))
                        continue

                    try:
                        text = self.formats['_time'] % self.stamp
                    except TypeError:
                        text = '%s' % self.stamp
                    cf.write(ns2u(text))

                    for field in fields:
                        if field in entry:
                            fmt = self.formats[tag][field]
                            value = entry[field]
                            try:
                                text = fmt % value
                            except TypeError:
                                text = '\t%s' % value
                            cf.write(ns2u(text))

                        else:  # field not in element
                            cf.write(u'\t')

                    cf.write(u'\n')

                try:
                    self.file.write(cf.getvalue())
                except ValueError as ex: #if self.file already closed then ValueError
                    console.terse("{0}\n".format(ex))

                cf.close()

    def never(self):
        """
        log never
        This if for manual logging by frame action
        """
        pass

    def once(self):
        """
        log once
        Good for logging paramters that don't change but want record
        """
        if self.stamp is None:
            self.log()

    def always(self):
        """
        log always
        Good for logging every time logger runs unconditionally
        """
        self.log()

    def streak(self):
        """
        log sequence of first field of first loggee only
        log elements in fifo order from sequence until empty
        """
        self.logStreak()

    def deck(self):
        """
        log deck
        log elements in fifo order from deck until empty
        """
        self.logDeck()


    def update(self):
        """
        log if updated
        logs once and then only if updated after first time
        """
        if self.stamp is None: #Always log at least once even if not updated
            self.log()
            return

        for loggee in self.loggees.values():
            if loggee.stamp is not None and loggee.stamp > self.stamp:  #any number is > None
                self.log()
                return  #first update triggers log once per cycle

    def change(self):
        """
        log if changed
        logs once and then only if changed after first time
        requires that self.prepare has been called otherwise fields in
        self.lasts won't match fields in log
        """
        if self.stamp is None: #Always log at least once even if not updated
            self.log()
            return

        change = False
        for tag, fields in self.fields.items():
            last = self.lasts[tag]  # get last Data object for each loggee
            loggee = self.loggees[tag]  # get loggee for tag

            try:
                for field in fields:
                    if not hasattr(last, field):  # was not present in prepare
                        if field in loggee:  # now present
                            change = True
                            setattr(last, field, loggee[field])

                    else:  # was present in prepare
                        if loggee[field] != getattr(last, field):
                            change = True
                            setattr(last, field, loggee[field])

            except AttributeError as ex: #
                console.terse("Warning: Log {0}, missing field"
                              " '{1}' for last value of loggee {2}\n".format(
                                  self.name, field, loggee.name))

            except KeyError as ex: #
                console.terse("Warning: Log {0}, missing field"
                              " '{1}' for loggee {2}\n".format(
                                  self.name, field, loggee.name))

        if change:
            self.log()

    def addLoggee(self, tag, loggee, fields=None):
        """
        Add a loggee at tag to .loggees
        Optional fields is list of field name strings to sub select fields in loggee
        """
        if self.stamp is None: #only add if not logged even once yet
            if tag ==  '_time':
                raise excepting.ResolveError("Bad loggee tag '_time'", self.name, loggee.name)
            if tag in self.loggees: #only add if not already there
                raise excepting.ResolveError("Duplicate tag", tag, loggee)
            self.loggees[tag] = loggee
            self.fields[tag] = fields if fields else []  # need tag to preserve order




# -*- coding: utf-8 -*-
"""
Unittests
"""
from __future__ import absolute_import, division, print_function

import sys

import unittest

import os
from collections import deque

from ioflo.aid.sixing import *
from ioflo.test import testing
from ioflo.aid import getConsole, odict

console = getConsole()


from ioflo.base import storing


def setUpModule():
    console.reinit(verbosity=console.Wordage.concise)

def tearDownModule():
    console.reinit(verbosity=console.Wordage.concise)

class BasicTestCase(unittest.TestCase):
    """
    Test Case
    """

    def setUp(self):
        """

        """
        console.reinit(verbosity=console.Wordage.profuse)

    def tearDown(self):
        """

        """
        console.reinit(verbosity=console.Wordage.concise)

    def testData(self):
        """
        Test Data Class
        """
        console.terse("{0}\n".format(self.testData.__doc__))

        data = storing.Data()
        show = data._show()
        self.assertEqual(show, 'Data: \n')
        self.assertEqual(repr(data), 'Data([])')
        data.x = 1
        data.y = 2
        data.a = 3
        # order preserving
        self.assertEqual(data.__dict__.keys(), ["x", "y", "a"])
        self.assertEqual(data.__dict__.values(), [1, 2, 3])
        self.assertEqual(data.__dict__.items(), [("x", 1),
                                                 ("y", 2),
                                                 ("a", 3)])
        show = data._show()
        self.assertEqual(show, 'Data: x=1 y=2 a=3\n')
        self.assertEqual(repr(data), "Data([('x', 1), ('y', 2), ('a', 3)])")
        stuff = data._sift()
        self.assertEqual(stuff, odict([("x", 1),
                                        ("y", 2),
                                        ("a", 3)]))
        self.assertEqual(stuff.keys(), data.__dict__.keys())

        stuff = data._sift(("y", "x"))
        self.assertEqual(stuff,odict([("y", 2),
                                        ("x", 1)]) )
        self.assertEqual(stuff.keys(), ["y", "x"])

        stuff = data._sift([])
        self.assertEqual(stuff,odict() )
        self.assertEqual(stuff.keys(), [])

        with self.assertRaises(AttributeError) as ex:
            data._b = 4
        self.assertEqual(repr(data), "Data([('x', 1), ('y', 2), ('a', 3)])")
        data._change([("z", 4), ("b", 5)])
        self.assertEqual(data.__dict__.keys(), ["x", "y", "a", "z", "b"])
        self.assertEqual(data.__dict__.values(), [1, 2, 3, 4, 5])
        self.assertEqual(data.__dict__.items(), [("x", 1),
                                                 ("y", 2),
                                                 ("a", 3),
                                                 ("z", 4),
                                                 ("b", 5)])

        items = [("x", 1),
                        ("y", 2),
                        ("a", 3),
                        ("z", 4),
                        ("b", 5)]
        stuff = odict(items)

        data = storing.Data(stuff)
        self.assertEqual(repr(data),
                         "Data([('x', 1), ('y', 2), ('a', 3), ('z', 4), ('b', 5)])")

        data = storing.Data(items)
        self.assertEqual(repr(data),
                         "Data([('x', 1), ('y', 2), ('a', 3), ('z', 4), ('b', 5)])")

        data = storing.Data(error=None)
        self.assertEqual(data.error, None)

        haul = odict([('error', None),
                      ('panSpeed', 50),
                      ('tiltSpeed', 75)])

        data._change(haul)
        self.assertEqual(data.panSpeed, haul['panSpeed'])
        self.assertEqual(data.tiltSpeed, haul['tiltSpeed'])
        # order preserving
        self.assertEqual(data.__dict__.keys(), ["error", "panSpeed", "tiltSpeed"])
        self.assertEqual(data.__dict__.values(), [None, 50, 75])
        self.assertEqual(data.__dict__.items(), [("error", None),
                                                 ("panSpeed", 50),
                                                 ("tiltSpeed", 75)])

        stuff = data._sift(haul.keys())
        self.assertEqual(haul, stuff)

        show = data._show()
        self.assertEqual(show, 'Data: error=None panSpeed=50 tiltSpeed=75\n')

        haul = odict([('panSpeed', 0),
                      ('tiltSpeed', 0)])

        data = storing.Data([('error', None)], haul)
        self.assertEqual(data.error, None)
        show = data._show()
        self.assertEqual(show, 'Data: error=None panSpeed=0 tiltSpeed=0\n')

    def testShare(self):
        """
        Test Share Class
        """
        console.terse("{0}\n".format(self.testShare.__doc__))

        share = storing.Share()
        self.assertEqual(share.name, '')
        self.assertEqual(share.store, None)
        self.assertEqual(share.stamp, None)
        self.assertIsInstance(share.data, storing.Data)
        self.assertEqual(share.data.__dict__, odict())
        self.assertEqual(share.value, None)
        self.assertEqual(share.truth, None)
        self.assertEqual(share.unit, None)
        self.assertEqual(share.owner, None)
        self.assertEqual(share.deck, storing.Deck([]))
        self.assertEqual(share.marks, odict([]))

        share = storing.Share()
        share.data.error = None
        self.assertEqual(share.data.error, None)
        self.assertEqual(share['error'], None)

        haul = odict([('error', None),
                      ('panSpeed', 50),
                      ('tiltSpeed', 75)])

        share.update(haul)
        self.assertEqual(share.data.panSpeed, haul['panSpeed'])
        self.assertEqual(share.data.tiltSpeed, haul['tiltSpeed'])
        # order preserving
        self.assertEqual(share.keys(), ["error", "panSpeed", "tiltSpeed"])
        self.assertEqual(share.values(), [None, 50, 75])
        self.assertEqual(share.items(), [("error", None),
                                                 ("panSpeed", 50),
                                                 ("tiltSpeed", 75)])

        stuff = share.sift(haul.keys())
        self.assertEqual(haul, stuff)

        show = share.data._show()
        self.assertEqual(show, 'Data: error=None panSpeed=50 tiltSpeed=75\n')

        haul = odict([('panSpeed', 0),
                      ('tiltSpeed', 0)])

        share = storing.Share().update([('error', None)], haul)
        self.assertEqual(share.data.error, None)
        show = share.data._show()
        self.assertEqual(show, 'Data: error=None panSpeed=0 tiltSpeed=0\n')

        show = share.show()  # need to make this more useful
        self.assertEqual(show, 'Name  Value None\nerror = None panSpeed = 0 tiltSpeed = 0\n')


    def testStore(self):
        """
        Test Store Class
        """
        console.terse("{0}\n".format(self.testStore.__doc__))
        storing.Store.Clear()  # clear registry of Store instance entries

        store = storing.Store()
        self.assertEqual(store.name, 'Store1')
        self.assertEqual(store.house, None)
        self.assertEqual(store.stamp, None)
        self.assertIsInstance(store.shares, storing.Node)
        self.assertIn("meta", store.shares)
        self.assertIn("realtime", store.shares)
        self.assertIn("time", store.shares)
        self.assertEqual(store.metaShr, storing.Node([]))
        self.assertIsInstance(store.timeShr, storing.Share)
        self.assertIsInstance(store.realTimeShr, storing.Share)
        self.assertIsInstance(store.dateTimeShr, storing.Share)
        for field in ("iso", "dt", "year", "month", "day", "hour", "minute", "second", "micro"):
            self.assertTrue(field in store.dateTimeShr)

        share = storing.Share(name='auto.depth')
        store.add(share).create(depth=10.0)
        self.assertIs(store.fetch('auto.depth'), share)
        self.assertIs(share.data.depth, 10.0)
        with  self.assertRaises(AttributeError) as  ex:
            share.data.value
        self.assertIs(share.value, None)

        path = 'auto.heading'
        store.create(path).create(heading=20.0)
        share = store.fetch(path)
        self.assertEqual(share.data.heading, 20.0)
        share.create(heading=25.0)
        self.assertEqual(share.data.heading, 20.0)

        shareA = store.create(path)
        self.assertIs(share, shareA)

        with self.assertRaises(ValueError) as  ex:
            store.add(storing.Share(name=path))

        shareA = storing.Share(name=path)
        store.change(shareA)
        self.assertIs(store.fetch(path), shareA)
        self.assertIsNot(store.fetch(path), share)
        shareA.value = 25.0
        self.assertEqual(shareA.value, 25.0)

        path = 'auto.speed'
        share = storing.Share(name=path)
        with self.assertRaises(ValueError) as  ex:
            store.change(share)
        store.add(share)
        self.assertIs(store.fetch(share.name), share)

        store.expose(valued=True)
        storing.Store.Clear()

    def testMark(self):
        """
        Test Mark Class
        """
        console.terse("{0}\n".format(self.testMark.__doc__))

        mark = storing.Mark()
        self.assertEqual(mark.stamp, None)
        self.assertEqual(mark.data, None)
        self.assertEqual(mark.used, None)

    def testDeck(self):
        """
        Test Deck Class
        """
        console.terse("{0}\n".format(self.testDeck.__doc__))

        deck = storing.Deck()
        self.assertIsInstance(deck, deque)
        self.assertEqual(storing.Deck.push, deque.append)
        self.assertEqual(storing.Deck.pull, deque.popleft)

        self.assertEqual(repr(deck), "Deck([])")

        self.assertEqual(len(deck), 0)
        self.assertEqual(deck.spew(), None)
        deck.gulp(None)
        self.assertEqual(len(deck), 0)


def runOne(test):
    '''
    Unittest Runner
    '''
    test = BasicTestCase(test)
    suite = unittest.TestSuite([test])
    unittest.TextTestRunner(verbosity=2).run(suite)

def runSome():
    """ Unittest runner """
    tests =  []
    names = [
                'testData',
                'testShare',
                'testStore',
                'testMark',
                'testDeck',
            ]
    tests.extend(map(BasicTestCase, names))
    suite = unittest.TestSuite(tests)
    unittest.TextTestRunner(verbosity=2).run(suite)

def runAll():
    """ Unittest runner """
    suite = unittest.TestSuite()
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(BasicTestCase))
    unittest.TextTestRunner(verbosity=2).run(suite)

if __name__ == '__main__' and __package__ is None:

    #runAll() #run all unittests

    runSome()#only run some

    #runOne('testBasic')







# -*- coding: utf-8 -*-
"""
Unittests
"""
from __future__ import absolute_import, division, print_function

import sys

import unittest

import os

from binascii import hexlify
import time

# Import ioflo libs
from ioflo.aid.sixing import *
from ioflo.aid.byting import hexify, bytify, unbytify, packify, unpackify
from ioflo.aid.timing import Timer, StoreTimer, Stamper
from ioflo.aid import getConsole

console = getConsole()

from ioflo.aio.proto import stacking, devicing, packeting

def setUpModule():
    console.reinit(verbosity=console.Wordage.concise)

def tearDownModule():
    console.reinit(verbosity=console.Wordage.concise)

class BasicTestCase(unittest.TestCase):
    """
    Test Case
    """

    def setUp(self):
        """

        """
        console.reinit(verbosity=console.Wordage.profuse)

    def tearDown(self):
        """

        """
        console.reinit(verbosity=console.Wordage.concise)

    def testStack(self):
        """
        Test Stack class
        """
        console.terse("{0}\n".format(self.testStack.__doc__))

        stack = stacking.Stack()
        self.assertIs(stack.local.stack, stack)
        self.assertEqual(stack.name, "Device1")
        self.assertEqual(stack.local.name, stack.name)
        self.assertEqual(stack.uid, 1)
        self.assertEqual(stack.local.uid, stack.uid)
        self.assertEqual(stack.ha, '')
        self.assertEqual(stack.aha, '')
        self.assertEqual(stack.local.ha, stack.ha)
        self.assertEqual(stack.kind, None)
        self.assertEqual(stack.local.kind, stack.kind)
        self.assertEqual(stack.handler, None)


    def testRemoteStack(self):
        """
        Test RemoteStack class
        """
        console.terse("{0}\n".format(self.testRemoteStack.__doc__))

        stack = stacking.RemoteStack()
        self.assertIs(stack.local.stack, stack)
        self.assertEqual(stack.name, "Device1")
        self.assertEqual(stack.local.name, stack.name)
        self.assertEqual(stack.uid, 1)
        self.assertEqual(stack.local.uid, stack.uid)
        self.assertEqual(stack.ha, '')
        self.assertEqual(stack.aha, '')
        self.assertEqual(stack.local.ha, stack.ha)
        self.assertEqual(stack.kind, None)
        self.assertEqual(stack.local.kind, stack.kind)
        self.assertEqual(stack.handler, None)

        self.assertEqual(len(stack.remotes), 0)
        self.assertIs(stack.remotes, stack.uidRemotes)
        self.assertEqual(len(stack.nameRemotes), 0)
        self.assertEqual(len(stack.haRemotes), 0)
        self.assertEqual(len(stack.stats), 0)

        remote = devicing.RemoteDevice(stack=stack, ha='hello')
        self.assertEqual(remote.uid, 2)
        self.assertEqual(remote.name, 'Device2')
        self.assertEqual(remote.ha, 'hello')
        stack.addRemote(remote)
        self.assertEqual(len(stack.remotes), 1)
        self.assertEqual(len(stack.nameRemotes), 1)
        self.assertEqual(len(stack.haRemotes), 1)
        self.assertIs(remote, stack.remotes[remote.uid])
        self.assertIs(remote, stack.nameRemotes[remote.name])
        self.assertIs(remote, stack.haRemotes[remote.ha])

        self.assertRaises(ValueError, stack.addRemote, remote)

        newUid = 3
        stack.moveRemote(remote, newUid)
        self.assertEqual(remote.uid, newUid)
        self.assertIs(remote, stack.remotes[newUid])

        newName = 'DeviceBlah'
        stack.renameRemote(remote, newName)
        self.assertEqual(remote.name, newName)
        self.assertIs(remote, stack.nameRemotes[newName])

        newHa = 'HelloWorld'
        stack.rehaRemote(remote, newHa)
        self.assertEqual(remote.ha, newHa)
        self.assertIs(remote, stack.haRemotes[newHa])

        stack.removeRemote(remote)
        self.assertEqual(len(stack.remotes), 0)
        self.assertEqual(len(stack.nameRemotes), 0)
        self.assertEqual(len(stack.haRemotes), 0)

    def testUdpStack(self):
        """
        Test UdpStack class
        """
        console.terse("{0}\n".format(self.testUdpStack.__doc__))

        stack = stacking.UdpStack()
        self.assertIsInstance(stack.local, devicing.IpLocalDevice)
        self.assertEqual(stack.local.uid, 1)
        self.assertEqual(stack.local.name, "Device{0}".format(stack.local.uid))
        self.assertEqual(stack.local.ha, ('127.0.0.1', stacking.UdpStack.Port))
        self.assertEqual(stack.local.kind, None)
        self.assertEqual(stack.aha, ('0.0.0.0', stacking.UdpStack.Port))
        stack.close()

        ha = ('127.0.0.1', 8000)
        stack = stacking.UdpStack(ha=ha)
        self.assertIsInstance(stack.local, devicing.IpLocalDevice)
        self.assertEqual(stack.local.uid, 1)
        self.assertEqual(stack.local.name, "Device{0}".format(stack.local.uid))
        self.assertEqual(stack.local.ha, ha)
        self.assertEqual(stack.local.kind, None)
        self.assertEqual(stack.aha, ha)
        stack.close()

    def testUdpStacks(self):
        """
        Test two stacks sending packets
        """
        console.terse("{0}\n".format(self.testUdpStacks.__doc__))

        stamper = Stamper()

        alpha = stacking.UdpStack(name='alpha',
                               port=8000)
        self.assertEqual(alpha.name, "alpha")
        self.assertEqual(alpha.ha, ('127.0.0.1', 8000))
        self.assertEqual(alpha.aha, ('0.0.0.0', 8000))
        self.assertIs(alpha.handler.opened, True)
        self.assertEqual(len(alpha.remotes), 0)

        remote = devicing.IpRemoteDevice(stack=alpha,
                                       name='BetaRemote',
                                       ha=('localhost', 8002))
        self.assertEqual(remote.uid, 2)
        self.assertEqual(remote.name, 'BetaRemote')
        self.assertEqual(remote.ha, ('127.0.0.1', 8002))
        alpha.addRemote(remote)


        beta = stacking.UdpStack(name='beta',
                              port=8002)
        self.assertEqual(beta.name, "beta")
        self.assertEqual(beta.ha, ('127.0.0.1', 8002))
        self.assertEqual(beta.aha, ('0.0.0.0', 8002))
        self.assertIs(beta.handler.opened, True)
        self.assertEqual(len(beta.remotes), 0)

        remote = devicing.IpRemoteDevice(stack=beta,
                                       name='AlphaRemote',
                                       ha=('localhost', 8000))
        self.assertEqual(remote.uid, 2)
        self.assertEqual(remote.name, 'AlphaRemote')
        self.assertEqual(remote.ha, ('127.0.0.1', 8000))
        beta.addRemote(remote)

        msg2beta = "Hello beta it's me alpha."
        alpha.message(msg2beta)

        msg2alpha = "Hi alpha from beta."
        beta.message(msg2alpha)

        #timer = Timer(duration=0.5)
        while (alpha.txMsgs or alpha.txPkts or alpha.rxPkts or alpha.rxMsgs
                or beta.txMsgs or  beta.txPkts or beta.rxPkts or beta.rxMsgs):
            alpha.serviceAllTx()
            beta.serviceAllTx()
            time.sleep(0.1)
            alpha.serviceAllRx()
            beta.serviceAllRx()
            stamper.advanceStamp(0.1)

        self.assertEqual(len(alpha.rxMsgs), 0)
        self.assertEqual(len(beta.rxMsgs), 0)
        self.assertEqual(alpha.stats['msg_received'], 1)
        self.assertEqual(beta.stats['msg_received'], 1)

        msg2beta = "Whats up beta?"
        packet = packeting.Packet(packed=msg2beta.encode('ascii'))
        alpha.transmit(packet)

        msg2alpha = "Just hanging alpha."
        packet = packeting.Packet(packed=msg2alpha.encode('ascii'))
        beta.transmit(packet)

        while (alpha.txMsgs or alpha.txPkts or alpha.rxPkts or alpha.rxMsgs
                or beta.txMsgs or  beta.txPkts or beta.rxPkts or beta.rxMsgs):
            alpha.serviceAllTx()
            beta.serviceAllTx()
            time.sleep(0.1)
            alpha.serviceAllRx()
            beta.serviceAllRx()
            stamper.advanceStamp(0.1)

        self.assertEqual(len(alpha.rxMsgs), 0)
        self.assertEqual(len(beta.rxMsgs), 0)
        self.assertEqual(alpha.stats['msg_received'], 2)
        self.assertEqual(beta.stats['msg_received'], 2)

        alpha.close()
        beta.close()

        alpha.reopen()
        beta.reopen()

        alpha.close()
        beta.close()

    def testTcpServerStack(self):
        """
        Test TcpServerStack class
        """
        console.terse("{0}\n".format(self.testTcpServerStack.__doc__))

        stack = stacking.TcpServerStack()
        self.assertIsInstance(stack.local, devicing.IpLocalDevice)
        self.assertEqual(stack.local.uid, 1)
        self.assertEqual(stack.local.name, "Device{0}".format(stack.local.uid))
        self.assertEqual(stack.local.ha, ('127.0.0.1', stacking.TcpServerStack.Port))
        self.assertEqual(stack.local.kind, None)
        self.assertEqual(stack.aha, ('0.0.0.0', stacking.UdpStack.Port))
        self.assertEqual(stack.eha, ('127.0.0.1', stacking.TcpServerStack.Port))
        stack.close()

        ha = ('127.0.0.1', 9000)
        stack = stacking.TcpServerStack(ha=ha)
        self.assertIsInstance(stack.local, devicing.IpLocalDevice)
        self.assertEqual(stack.local.uid, 1)
        self.assertEqual(stack.local.name, "Device{0}".format(stack.local.uid))
        self.assertEqual(stack.local.ha, ha)
        self.assertEqual(stack.local.kind, None)
        self.assertEqual(stack.aha, ha)
        self.assertEqual(stack.eha, ha)
        stack.close()

    def testTcpClientStack(self):
        """
        Test TcpClientStack class
        """
        console.terse("{0}\n".format(self.testTcpClientStack.__doc__))

        stack = stacking.TcpClientStack()
        self.assertIsInstance(stack.local, devicing.IpLocalDevice)
        self.assertEqual(stack.local.uid, 1)
        self.assertEqual(stack.local.name, "Device{0}".format(stack.local.uid))
        self.assertEqual(stack.local.ha, (None, None))  # not connected yet
        self.assertEqual(stack.local.kind, None)
        self.assertEqual(stack.aha, ('127.0.0.1', stacking.TcpServerStack.Port))
        self.assertEqual(stack.aha, stack.ha)
        self.assertEqual(stack.ha, stack.remote.ha)
        self.assertEqual(stack.aha, ('127.0.0.1', stacking.TcpServerStack.Port))
        self.assertEqual(stack.remote.ha, ('127.0.0.1', stacking.TcpServerStack.Port))
        stack.close()

        ha = ('127.0.0.1', 9000)
        stack = stacking.TcpClientStack(ha=ha)
        self.assertIsInstance(stack.local, devicing.IpLocalDevice)
        self.assertEqual(stack.local.uid, 1)
        self.assertEqual(stack.local.name, "Device{0}".format(stack.local.uid))
        self.assertEqual(stack.local.ha, (None, None))
        self.assertEqual(stack.local.kind, None)
        self.assertEqual(stack.ha, ha)
        self.assertEqual(stack.aha, stack.ha)
        self.assertEqual(stack.ha, stack.remote.ha)
        self.assertEqual(stack.remote.ha, ('127.0.0.1', 9000))
        self.assertEqual(stack.aha, ha)
        stack.close()



def runOne(test):
    '''
    Unittest Runner
    '''
    test = BasicTestCase(test)
    suite = unittest.TestSuite([test])
    unittest.TextTestRunner(verbosity=2).run(suite)

def runSome():
    """ Unittest runner """
    tests =  []
    names = [
             'testStack',
             'testRemoteStack',
             'testUdpStack',
             'testUdpStacks',
             'testTcpServerStack',
             'testTcpClientStack',
            ]
    tests.extend(map(BasicTestCase, names))
    suite = unittest.TestSuite(tests)
    unittest.TextTestRunner(verbosity=2).run(suite)

def runAll():
    """ Unittest runner """
    suite = unittest.TestSuite()
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(BasicTestCase))
    unittest.TextTestRunner(verbosity=2).run(suite)

if __name__ == '__main__' and __package__ is None:

    #runAll() #run all unittests

    runSome()#only run some

    #runOne('testPart')






"""
checking.py utility functions for crc or chechsum type checking
"""
from __future__ import absolute_import, division, print_function

import sys
import os
import struct

# Import ioflo libs
from .sixing import *
from .consoling import getConsole

console = getConsole()


def crc16(inpkt):
    """ Returns 16 bit crc or inpkt packed binary string
        compatible with ANSI 709.1 and 852
        inpkt is bytes in python3 or str in python2
        needs struct module
    """
    inpkt = bytearray(inpkt)
    poly = 0x1021  # Generator Polynomial
    crc = 0xffff
    for element in inpkt :
        i = 0
        #byte = ord(element)
        byte = element
        while i < 8 :
            crcbit = 0x0
            if (crc & 0x8000):
                crcbit = 0x01
            databit = 0x0
            if (byte & 0x80):
                databit = 0x01
            crc = crc << 1
            crc = crc & 0xffff
            if (crcbit != databit):
                crc = crc ^ poly
            byte = byte << 1
            byte = byte & 0x00ff
            i += 1
    crc = crc ^ 0xffff
    return struct.pack("!H",crc )

def crc64(inpkt) :
    """ Returns 64 bit crc of inpkt binary packed string inpkt
        inpkt is bytes in python3 or str in python2
        returns tuple of two 32 bit numbers for top and bottom of 64 bit crc
    """
    inpkt = bytearray(inpkt)
    polytop = 0x42f0e1eb
    polybot = 0xa9ea3693
    crctop  = 0xffffffff
    crcbot  = 0xffffffff
    for element in inpkt :
        i = 0
        #byte = ord(element)
        byte = element
        while i < 8 :
            topbit = 0x0
            if (crctop & 0x80000000):
                topbit = 0x01
            databit = 0x0
            if (byte & 0x80):
                databit = 0x01
            crctop = crctop << 1
            crctop = crctop & 0xffffffff
            botbit = 0x0
            if (crcbot & 0x80000000):
                botbit = 0x01
            crctop = crctop | botbit
            crcbot = crcbot << 1
            crcbot = crcbot & 0xffffffff
            if (topbit != databit):
                crctop = crctop ^ polytop
                crcbot = crcbot ^ polybot
            byte = byte << 1
            byte = byte & 0x00ff
            i += 1
    crctop = crctop ^ 0xffffffff
    crcbot = crcbot ^ 0xffffffff
    return (crctop, crcbot)




# -*- coding: utf-8 -*-
"""
Unit Test Template
"""
from __future__ import absolute_import, division, print_function

import sys
import datetime
import unittest

import os
import time

from ioflo.aid.sixing import *
from ioflo.aid.odicting import odict
from ioflo.test import testing
from ioflo.aid.consoling import getConsole
from ioflo.aid.timing import iso8601, tuuid

console = getConsole()

from ioflo.aid import eventing


def setUpModule():
    console.reinit(verbosity=console.Wordage.concise)

def tearDownModule():
    pass


class BasicTestCase(unittest.TestCase):
    """
    Example TestCase
    """

    def setUp(self):
        """
        Call super if override so House Framer and Frame are setup correctly
        """
        super(BasicTestCase, self).setUp()

    def tearDown(self):
        """
        Call super if override so House Framer and Frame are torn down correctly
        """
        super(BasicTestCase, self).tearDown()

    def testTagify(self):
        """
        Test tagify function
        """
        console.terse("{0}\n".format(self.testTagify.__doc__))
        console.reinit(verbosity=console.Wordage.profuse)

        tag = eventing.tagify()
        self.assertEqual(tag, u'')

        tag = eventing.tagify(head='exchange')
        self.assertEqual(tag, 'exchange')

        tag = eventing.tagify(head=['exchange', 'trade'])
        self.assertEqual(tag, 'exchange.trade')

        tag = eventing.tagify(head='exchange', tail='completed')
        self.assertEqual(tag, 'exchange.completed')

        tag = eventing.tagify(head='exchange', tail=['process', 'started'])
        self.assertEqual(tag, 'exchange.process.started')

        tag = eventing.tagify(head=['exchange', 'trade'], tail=['process', 'started'])
        self.assertEqual(tag, 'exchange.trade.process.started')

        tag = eventing.tagify(head=['exchange', 'trade'], tail='completed')
        self.assertEqual(tag, 'exchange.trade.completed')

        tag = eventing.tagify(head='exchange', tail=['process', 'started'], sep='/')
        self.assertEqual(tag, 'exchange/process/started')

        tag = eventing.tagify(tail=['process', 'started'])
        self.assertEqual(tag, 'process.started')

        console.reinit(verbosity=console.Wordage.concise)

    def testEventify(self):
        """
        Test eventify function
        """
        console.terse("{0}\n".format(self.testEventify.__doc__))
        console.reinit(verbosity=console.Wordage.profuse)

        dt = datetime.datetime.utcnow()
        stamp = dt.isoformat()
        time.sleep(0.01)

        event = eventing.eventify('hello')
        self.assertEqual(event['tag'], 'hello')
        self.assertEqual(event['data'], {})
        #"YYYY-MM-DDTHH:MM:SS.mmmmmm"
        tdt = datetime.datetime.strptime(event['stamp'], "%Y-%m-%dT%H:%M:%S.%f")
        self.assertGreater(tdt, dt)

        event = eventing.eventify(tag=eventing.tagify(head='exchange', tail='started'),
                                stamp=stamp)
        self.assertEqual(event['tag'], 'exchange.started' )
        self.assertEqual(event['stamp'], stamp )

        event = eventing.eventify(tag=eventing.tagify(tail='started', head='exchange'),
                                stamp=stamp,
                                data = odict(name='John'))
        self.assertEqual(event['tag'], 'exchange.started')
        self.assertEqual(event['stamp'], stamp)
        self.assertEqual(event['data'], {'name':  'John',})

        stamp = '2015-08-10T19:26:47.194736'
        event = eventing.eventify(tag='process.started', stamp=stamp, data={'name': 'Jill',})
        self.assertEqual(event, {'tag': 'process.started',
                                 'stamp': '2015-08-10T19:26:47.194736',
                                 'data': {'name': 'Jill',},})

        event = eventing.eventify(tag="with uid", stamp=stamp, uid="abcde")
        self.assertEqual(event, {'data': {},
                                'stamp': '2015-08-10T19:26:47.194736',
                                'tag': 'with uid',
                                'uid': 'abcde'})

        console.reinit(verbosity=console.Wordage.concise)

    def testEventize(self):
        """
        Test eventize function
        """
        console.terse("{0}\n".format(self.testEventize.__doc__))
        console.reinit(verbosity=console.Wordage.profuse)


        stamp = iso8601()  # "YYYY-MM-DDTHH:MM:SS.mmmmmm"
        tuid = tuuid()  # "0000014ddf1f2f9c_5e36738"
        time.sleep(0.1)

        event = eventing.eventize('hello')
        self.assertEqual(event['tag'], 'hello')
        self.assertFalse('data' in event)
        self.assertFalse('stamp' in event)
        self.assertFalse('uid' in event)
        self.assertFalse('route' in event)

        event = eventing.eventize(tag=eventing.tagify(head='exchange', tail='started'),
                                  stamp=True,
                                  uid=True,
                                  data=True,
                                  route=odict([("src", (None, None, None)),
                                             ("dst", (None, None, None))]))


        self.assertEqual(event['tag'], 'exchange.started')
        self.assertTrue('data' in event)
        self.assertIsInstance(event["data"], odict)
        self.assertEqual(event['data'], odict([]))
        self.assertTrue('stamp' in event)
        self.assertIsInstance(event["stamp"], str)
        self.assertGreater(event['stamp'], stamp)
        self.assertTrue('uid' in event)
        self.assertIsInstance(event["uid"], str)
        self.assertGreater(event['uid'], tuid)
        self.assertTrue('route' in event)
        self.assertEqual(event['route'] ,odict([("src", (None, None, None)),
                                         ("dst", (None, None, None))]))

        event = eventing.eventize(tag=eventing.tagify(head='exchange', tail='started'),
                                  stamp=stamp,
                                  uid=tuid,
                                  data=odict(name="John"),
                                  route=odict([("src", (None, None, None)),
                                             ("dst", (None, None, None))]))


        self.assertEqual(event['tag'], 'exchange.started')
        self.assertEqual(event['data'], odict(name="John"))
        self.assertEqual(event['stamp'], stamp)
        self.assertEqual(event['uid'], tuid)
        self.assertEqual(event['route'] ,odict([("src", (None, None, None)),
                                         ("dst", (None, None, None))]))



        console.reinit(verbosity=console.Wordage.concise)


def runOne(test):
    '''
    Unittest Runner
    '''
    test = BasicTestCase(test)
    suite = unittest.TestSuite([test])
    unittest.TextTestRunner(verbosity=2).run(suite)

def runSome():
    """ Unittest runner """
    tests =  []
    names = [
             'testTagify',
             'testEventify',
             'testEventize',
            ]
    tests.extend(map(BasicTestCase, names))
    suite = unittest.TestSuite(tests)
    unittest.TextTestRunner(verbosity=2).run(suite)

def runAll():
    """ Unittest runner """
    suite = unittest.TestSuite()
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(BasicTestCase))
    unittest.TextTestRunner(verbosity=2).run(suite)

if __name__ == '__main__' and __package__ is None:

    #console.reinit(verbosity=console.Wordage.concise)

    #runAll() #run all unittests

    runSome()#only run some

    #runOne('testBasic')





"""
filing.py file and serialization utility functions
"""
from __future__ import absolute_import, division, print_function

import sys
import os
import errno

try:
    import simplejson as json
except ImportError:
    import json

# Import ioflo libs
from .sixing import *
from .odicting import odict
from .consoling import getConsole

console = getConsole()


def ocfn(filename, openMode='r+', binary=False):
    """Atomically open or create file from filename.

       If file already exists, Then open file using openMode
       Else create file using write update mode If not binary Else
           write update binary mode
       Returns file object

       If binary Then If new file open with write update binary mode
    """
    try:
        newfd = os.open(filename, os.O_EXCL | os.O_CREAT | os.O_RDWR, 436) # 436 == octal 0664
        if not binary:
            newfile = os.fdopen(newfd,"w+")
        else:
            newfile = os.fdopen(newfd,"w+b")
    except OSError as ex:
        if ex.errno == errno.EEXIST:
            newfile = open(filename, openMode)
        else:
            raise
    return newfile

def dump(data, filepath):
    '''
    Write data as as type self.ext to filepath. json or .msgpack
    '''
    if ' ' in filepath:
        raise IOError("Invalid filepath '{0}' "
                                "contains space".format(filepath))

    root, ext = os.path.splitext(filepath)
    if ext == '.json':
        with ocfn(filepath, "w+") as f:
            json.dump(data, f, indent=2)
            f.flush()
            os.fsync(f.fileno())
    elif ext == '.msgpack':
        if not msgpack:
            raise IOError("Invalid filepath ext '{0}' "
                        "needs msgpack installed".format(filepath))
        with ocfn(filepath, "w+b", binary=True) as f:
            msgpack.dump(data, f)
            f.flush()
            os.fsync(f.fileno())
    else:
        raise IOError("Invalid filepath ext '{0}' "
                    "not '.json' or '.msgpack'".format(filepath))

def load(filepath):
    '''
    Return data read from filepath as dict
    file may be either json or msgpack given by extension .json or .msgpack
    Otherwise return None
    '''
    try:
        root, ext = os.path.splitext(filepath)
        if ext == '.json':
            with ocfn(filepath, "r") as f:
                it = json.load(f, object_pairs_hook=odict)
        elif ext == '.msgpack':
            if not msgpack:
                raise IOError("Invalid filepath ext '{0}' "
                            "needs msgpack installed".format(filepath))
            with ocfn(filepath, "rb", binary=True) as f:
                it = msgpack.load(f, object_pairs_hook=odict)
        else:
            it = None
    except EOFError:
        return None
    except ValueError:
        return None
    return it

def loadPickle(file=""):
    """
    Loads pickled object from file, returns object
    """
    if not file:
        raise ParameterError("No file to load from: {0}".format(file))

    with open(file,"r+") as f:
        p = pickle.Unpickler(f)
        it = p.load()

    return it

def dumpPickle(it=None, file=""):
    """
    Pickles it object to file
    """
    if not it:
        raise ParameterError("No object to dump: {0}".format(str(it)))

    if not file:
        raise ParameterError("No file to dump to: {0}".format(file))

    with open(file, "w+") as f:
        p = pickle.Pickler(f)
        p.dump(it)

def loadJson(filename=""):
    """ Loads json object from filename, returns unjsoned object"""
    if not filename:
        raise ParameterError("Empty filename to load.")

    with ocfn(filename) as f:
        try:
            it = json.load(f, object_pairs_hook=odict())
        except EOFError:
            return None
        except ValueError:
            return None
        return it

def dumpJson(it=None, filename="", indent=2):
    """Jsonifys it and dumps it to filename"""
    if not it:
        raise ValueError("No object to Dump: {0}".format(it))

    if not filename:
        raise ValueError("No file to Dump to: {0}".format(filename))

    with ocfn(filename, "w+") as f:
        json.dump(it, f, indent=2)
        f.flush()
        os.fsync(f.fileno())



# -*- coding: utf-8 -*-
"""
Unittests
"""
from __future__ import absolute_import, division, print_function

import sys

import unittest

import os

from ioflo.aid.sixing import *
from ioflo.test import testing
from ioflo.aid import getConsole, odict

console = getConsole()


from ioflo.base import globaling


def setUpModule():
    console.reinit(verbosity=console.Wordage.concise)

def tearDownModule():
    console.reinit(verbosity=console.Wordage.concise)

class BasicTestCase(unittest.TestCase):
    """
    Test Case
    """

    def setUp(self):
        """

        """
        console.reinit(verbosity=console.Wordage.profuse)

    def tearDown(self):
        """

        """
        console.reinit(verbosity=console.Wordage.concise)

    def testPointRegex(self):
        """
        Test Point Regex Instances
        """
        console.terse("{0}\n".format(self.testPointRegex.__doc__))

        text = "-25.6x30.45y"
        match = globaling.REO_PointXY.findall(text)
        x, y = match[0]
        x = float(x)
        y = float(y)
        self.assertEqual(x, -25.6)
        self.assertEqual(y, 30.45)

        text = "-25.6X+30.45Y"
        match = globaling.REO_PointXY.findall(text)
        x, y = match[0]
        x = float(x)
        y = float(y)
        self.assertEqual(x, -25.6)
        self.assertEqual(y, 30.45)

        text = "-25.6x30.45Y"
        match = globaling.REO_PointXY.findall(text)
        x, y = match[0]
        x = float(x)
        y = float(y)
        self.assertEqual(x, -25.6)
        self.assertEqual(y, 30.45)

        text = "+25x-30y"
        match = globaling.REO_PointXY.findall(text)
        x, y = match[0]
        x = float(x)
        y = float(y)
        self.assertEqual(x, 25.0)
        self.assertEqual(y, -30.0)

        text = "25.x+30.y"
        match = globaling.REO_PointXY.findall(text)
        x, y = match[0]
        x = float(x)
        y = float(y)
        self.assertEqual(x, 25.0)
        self.assertEqual(y, 30.0)

        text = "-25.6n30.45e"
        match = globaling.REO_PointNE.findall(text)
        n, e = match[0]
        n = float(n)
        e = float(e)
        self.assertEqual(n, -25.6)
        self.assertEqual(e, 30.45)

        text = "-25.6N+30.45E"
        match = globaling.REO_PointNE.findall(text)
        n, e = match[0]
        n = float(n)
        e = float(e)
        self.assertEqual(n, -25.6)
        self.assertEqual(e, 30.45)

        text = "-25.6n30.45E"
        match = globaling.REO_PointNE.findall(text)
        n, e = match[0]
        n = float(n)
        e = float(e)
        self.assertEqual(n, -25.6)
        self.assertEqual(e, 30.45)

        text = "25n-30e"
        match = globaling.REO_PointNE.findall(text)
        n, e = match[0]
        n = float(n)
        e = float(e)
        self.assertEqual(n, 25.0)
        self.assertEqual(e, -30.0)

        text = "+25.n30.e"
        match = globaling.REO_PointNE.findall(text)
        n, e = match[0]
        n = float(n)
        e = float(e)
        self.assertEqual(n, 25.0)
        self.assertEqual(e, 30.0)

        text = "25.6f-30.45s"
        match = globaling.REO_PointFS.findall(text)
        f, s = match[0]
        f = float(f)
        s = float(s)
        self.assertEqual(f, 25.6)
        self.assertEqual(s, -30.45)

        text = "-25.6F+30.45S"
        match = globaling.REO_PointFS.findall(text)
        f, s = match[0]
        f = float(f)
        s = float(s)
        self.assertEqual(f, -25.6)
        self.assertEqual(s, 30.45)


        text = "25.6x-30.45y-15.4z"
        match = globaling.REO_PointXYZ.findall(text)
        x, y, z = match[0]
        x = float(x)
        y = float(y)
        z = float(z)
        self.assertEqual(x, 25.6)
        self.assertEqual(y, -30.45)
        self.assertEqual(z, -15.4)

        text = "-25.6X+30.45Y15.4Z"
        match = globaling.REO_PointXYZ.findall(text)
        x, y, z = match[0]
        x = float(x)
        y = float(y)
        z = float(z)
        self.assertEqual(x, -25.6)
        self.assertEqual(y, 30.45)
        self.assertEqual(z, 15.4)

        text = "25.6n-30.45e-15.4d"
        match = globaling.REO_PointNED.findall(text)
        n, e, d = match[0]
        n = float(n)
        e = float(e)
        d = float(d)
        self.assertEqual(n, 25.6)
        self.assertEqual(e, -30.45)
        self.assertEqual(d, -15.4)

        text = "-25.6N+30.45E15.4D"
        match = globaling.REO_PointNED.findall(text)
        n, e, d  = match[0]
        n = float(n)
        e = float(e)
        d = float(d)
        self.assertEqual(n, -25.6)
        self.assertEqual(e, 30.45)
        self.assertEqual(d, 15.4)

        text = "25.6f-30.45s-15.4b"
        match = globaling.REO_PointFSB.findall(text)
        f, s, b = match[0]
        f = float(f)
        s = float(s)
        b = float(b)
        self.assertEqual(f, 25.6)
        self.assertEqual(s, -30.45)
        self.assertEqual(b, -15.4)

        text = "-25.6F+30.45S15.4B"
        match = globaling.REO_PointFSB.findall(text)
        f, s, b = match[0]
        f = float(f)
        s = float(s)
        b = float(b)
        self.assertEqual(f, -25.6)
        self.assertEqual(s, 30.45)
        self.assertEqual(b, 15.4)


def runOne(test):
    '''
    Unittest Runner
    '''
    test = BasicTestCase(test)
    suite = unittest.TestSuite([test])
    unittest.TextTestRunner(verbosity=2).run(suite)

def runSome():
    """ Unittest runner """
    tests =  []
    names = [
                'testPointRegex',
            ]
    tests.extend(map(BasicTestCase, names))
    suite = unittest.TestSuite(tests)
    unittest.TextTestRunner(verbosity=2).run(suite)

def runAll():
    """ Unittest runner """
    suite = unittest.TestSuite()
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(BasicTestCase))
    unittest.TextTestRunner(verbosity=2).run(suite)

if __name__ == '__main__' and __package__ is None:

    #runAll() #run all unittests

    runSome()#only run some

    #runOne('testBasic')







# -*- coding: utf-8 -*-
"""
Unit Test Template
"""

import sys
import unittest

import os
import time

from ioflo.aid.sixing import *
from ioflo.aid import odict
from ioflo.test import testing
from ioflo.base import globaling
from ioflo.base import Data, Deck

from ioflo.aid.consoling import getConsole

from ioflo.base import storing
from ioflo.base import logging

console = getConsole()


def setUpModule():
    console.reinit(verbosity=console.Wordage.profuse)

def tearDownModule():
    console.reinit(verbosity=console.Wordage.concise)


class LoggerTestCase(testing.LoggerIofloTestCase):
    """
    Example TestCase
    """

    def setUp(self):
        """
        Call super if override so House Framer and Frame are setup correctly
        """
        super(LoggerTestCase, self).setUp()

    def tearDown(self):
        """
        Call super if override so House Framer and Frame are torn down correctly
        """
        super(LoggerTestCase, self).tearDown()


    def testLogger(self):
        """
        Test creating a logger with a log and loggees
        """
        console.terse("{0}\n".format(self.testLogger.__doc__))
        self.assertEqual(self.house.store, self.store)
        self.assertIsInstance(self.logger, logging.Logger)
        self.assertEqual(self.logger.prefix, '/tmp/log/ioflo')
        self.assertTrue(self.logger.runner)  # runner generator is made when logger created

        heading = self.store.create('pose.heading').create(value = 0.0)
        position = self.store.create('pose.position').create([("north", 10.0), ("east", 5.0)])

        log = logging.Log(name='test',
                          store=self.store,
                          kind='text',
                          baseFilename='',
                          rule=globaling.ALWAYS)

        self.assertEqual(log.baseFilename, log.name)
        self.assertEqual(log.path, '')
        self.assertEqual(log.file, None)
        self.assertEqual(log.kind, 'text')

        self.assertEqual(log.rule, globaling.ALWAYS)
        self.assertEqual(log.action, log.always)

        log.addLoggee(tag = 'heading', loggee = 'pose.heading')
        log.addLoggee(tag = 'pos', loggee = 'pose.position')
        log.resolve()

        self.logger.addLog(log)
        self.logger.reopen()  # this updates paths on all logs
        self.assertTrue(self.logger.path.startswith('/tmp/log/ioflo/HouseTest/LoggerTest_'))
        self.assertTrue(log.path.startswith(self.logger.path))
        self.assertTrue(log.path.endswith(log.baseFilename + '.txt'))
        self.assertTrue(log.file)

        log.prepare()
        self.assertEqual(log.formats, {'_time': '%s',
                                        'heading': odict([('value', '\t%s')]),
                                        'pos': odict([('north', '\t%s'), ('east', '\t%s')])})
        #self.assertEqual(log.formats, {'_time': '%0.4f',
                                        #'heading': odict([('value', '\t%0.4f')]),
                                        #'pos': odict([('north', '\t%0.4f'), ('east', '\t%0.4f')])})

        self.house.store.changeStamp(0.0)

        log() #log
        for i in range(20):
            self.store.advanceStamp(0.125)
            if i == 5:
                heading.value += 0.0
                position.data.north += 0.0
                position.data.east -= 0.0
            elif i == 10:
                pass
            else:
                heading.value = float(i)
                position.data.north += 2.0
                position.data.east -= 1.5

            log() #log

        log.flush()
        log.file.seek(0)
        line = log.file.readline()
        self.assertEqual(line, 'text\tAlways\ttest\n')
        line = log.file.readline()
        self.assertEqual(line, '_time\theading\tpos.north\tpos.east\n')
        lines = log.file.readlines()
        self.assertEqual(lines, ['0.0\t0.0\t10.0\t5.0\n',
                                '0.125\t0.0\t12.0\t3.5\n',
                                '0.25\t1.0\t14.0\t2.0\n',
                                '0.375\t2.0\t16.0\t0.5\n',
                                '0.5\t3.0\t18.0\t-1.0\n',
                                '0.625\t4.0\t20.0\t-2.5\n',
                                '0.75\t4.0\t20.0\t-2.5\n',
                                '0.875\t6.0\t22.0\t-4.0\n',
                                '1.0\t7.0\t24.0\t-5.5\n',
                                '1.125\t8.0\t26.0\t-7.0\n',
                                '1.25\t9.0\t28.0\t-8.5\n',
                                '1.375\t9.0\t28.0\t-8.5\n',
                                '1.5\t11.0\t30.0\t-10.0\n',
                                '1.625\t12.0\t32.0\t-11.5\n',
                                '1.75\t13.0\t34.0\t-13.0\n',
                                '1.875\t14.0\t36.0\t-14.5\n',
                                '2.0\t15.0\t38.0\t-16.0\n',
                                '2.125\t16.0\t40.0\t-17.5\n',
                                '2.25\t17.0\t42.0\t-19.0\n',
                                '2.375\t18.0\t44.0\t-20.5\n',
                                '2.5\t19.0\t46.0\t-22.0\n'])

        log.close()

        try:
            os.remove(log.path)  # remove to clean up
        except OSError as ex:
            pass



    def testLogAlways(self):
        """
        Test creating a logger with a log and loggees
        """
        console.terse("{0}\n".format(self.testLogAlways.__doc__))

        self.assertEqual(self.house.store, self.store)
        self.assertIsInstance(self.logger, logging.Logger)
        self.assertEqual(self.logger.prefix, '/tmp/log/ioflo')
        self.assertTrue(self.logger.runner)  # runner generator is made when logger created

        heading = self.store.create('pose.heading').create(value = 0.0)
        position = self.store.create('pose.position').create([("north", 10.0), ("east", 5.0)])

        log = logging.Log(name='test',
                          store=self.store,
                          kind='text',
                          baseFileName='',
                          rule=globaling.ALWAYS)

        self.assertEqual(log.baseFilename, log.name)
        self.assertEqual(log.path, '')
        self.assertEqual(log.file, None)
        self.assertEqual(log.kind, 'text')

        self.assertEqual(log.rule, globaling.ALWAYS)
        self.assertEqual(log.action, log.always)

        log.addLoggee(tag = 'heading', loggee = 'pose.heading')
        log.addLoggee(tag = 'pos', loggee = 'pose.position')

        self.logger.addLog(log)
        self.logger.resolve()  # resolves logs as well


        self.house.store.changeStamp(0.0)
        self.assertIs(log.stamp, None)
        status = self.logger.runner.send(globaling.START)  # reopens prepares and logs once
        self.assertTrue(self.logger.path.startswith('/tmp/log/ioflo/HouseTest/LoggerTest_'))
        self.assertTrue(log.path.startswith(self.logger.path))
        self.assertTrue(log.path.endswith(log.baseFilename + '.txt'))
        self.assertTrue(log.file)
        self.assertEqual(log.formats, {'_time': '%s',
                                        'heading': odict([('value', '\t%s')]),
                                        'pos': odict([('north', '\t%s'), ('east', '\t%s')])})

        #self.assertEqual(log.formats, {'_time': '%0.4f',
                                        #'heading': odict([('value', '\t%0.4f')]),
                                        #'pos': odict([('north', '\t%0.4f'), ('east', '\t%0.4f')])})

        for i in range(20):
            self.store.advanceStamp(0.125)
            if i == 5:
                heading.value += 0.0
                position.data.north += 0.0
                position.data.east -= 0.0
            elif i == 10:
                pass
            else:
                heading.value = float(i)
                position.data.north += 2.0
                position.data.east -= 1.5

            status = self.logger.runner.send(globaling.RUN)

        self.store.advanceStamp(0.125)
        status = self.logger.runner.send(globaling.STOP)  # logs once and closes logs

        log.reopen()
        log.file.seek(0)  # reopen appends so seek back to start
        line = log.file.readline()
        self.assertEqual(line, 'text\tAlways\ttest\n')
        line = log.file.readline()
        self.assertEqual(line, '_time\theading\tpos.north\tpos.east\n')
        lines = log.file.readlines()
        self.assertEqual(lines, ['0.0\t0.0\t10.0\t5.0\n',
                                    '0.125\t0.0\t12.0\t3.5\n',
                                    '0.25\t1.0\t14.0\t2.0\n',
                                    '0.375\t2.0\t16.0\t0.5\n',
                                    '0.5\t3.0\t18.0\t-1.0\n',
                                    '0.625\t4.0\t20.0\t-2.5\n',
                                    '0.75\t4.0\t20.0\t-2.5\n',
                                    '0.875\t6.0\t22.0\t-4.0\n',
                                    '1.0\t7.0\t24.0\t-5.5\n',
                                    '1.125\t8.0\t26.0\t-7.0\n',
                                    '1.25\t9.0\t28.0\t-8.5\n',
                                    '1.375\t9.0\t28.0\t-8.5\n',
                                    '1.5\t11.0\t30.0\t-10.0\n',
                                    '1.625\t12.0\t32.0\t-11.5\n',
                                    '1.75\t13.0\t34.0\t-13.0\n',
                                    '1.875\t14.0\t36.0\t-14.5\n',
                                    '2.0\t15.0\t38.0\t-16.0\n',
                                    '2.125\t16.0\t40.0\t-17.5\n',
                                    '2.25\t17.0\t42.0\t-19.0\n',
                                    '2.375\t18.0\t44.0\t-20.5\n',
                                    '2.5\t19.0\t46.0\t-22.0\n',
                                    '2.625\t19.0\t46.0\t-22.0\n'])

        self.assertIsNot(log.stamp, None)
        self.store.advanceStamp(0.125)
        status = self.logger.runner.send(globaling.START)  # logs once no headers since log stamp not none
        self.store.advanceStamp(0.125)
        heading.value += 5.0
        status = self.logger.runner.send(globaling.RUN)  # logs once
        self.store.advanceStamp(0.125)
        status = self.logger.runner.send(globaling.STOP)  # logs once and closes logs

        log.reopen()
        log.file.seek(0)  # reopen appends so seek back to start
        lines = log.file.readlines()
        self.assertEqual(lines, ['text\tAlways\ttest\n',
                                '_time\theading\tpos.north\tpos.east\n',
                                '0.0\t0.0\t10.0\t5.0\n',
                                '0.125\t0.0\t12.0\t3.5\n',
                                '0.25\t1.0\t14.0\t2.0\n',
                                '0.375\t2.0\t16.0\t0.5\n',
                                '0.5\t3.0\t18.0\t-1.0\n',
                                '0.625\t4.0\t20.0\t-2.5\n',
                                '0.75\t4.0\t20.0\t-2.5\n',
                                '0.875\t6.0\t22.0\t-4.0\n',
                                '1.0\t7.0\t24.0\t-5.5\n',
                                '1.125\t8.0\t26.0\t-7.0\n',
                                '1.25\t9.0\t28.0\t-8.5\n',
                                '1.375\t9.0\t28.0\t-8.5\n',
                                '1.5\t11.0\t30.0\t-10.0\n',
                                '1.625\t12.0\t32.0\t-11.5\n',
                                '1.75\t13.0\t34.0\t-13.0\n',
                                '1.875\t14.0\t36.0\t-14.5\n',
                                '2.0\t15.0\t38.0\t-16.0\n',
                                '2.125\t16.0\t40.0\t-17.5\n',
                                '2.25\t17.0\t42.0\t-19.0\n',
                                '2.375\t18.0\t44.0\t-20.5\n',
                                '2.5\t19.0\t46.0\t-22.0\n',
                                '2.625\t19.0\t46.0\t-22.0\n',
                                '2.75\t19.0\t46.0\t-22.0\n',
                                '2.875\t24.0\t46.0\t-22.0\n',
                                '3.0\t24.0\t46.0\t-22.0\n'])
        log.file.close()

        try:
            os.remove(log.path)  # remove to clean up
        except OSError as ex:
            pass


    def testLogOnce(self):
        """
        Test log with once rule
        """
        console.terse("{0}\n".format(self.testLogOnce.__doc__))

        self.assertEqual(self.house.store, self.store)
        self.assertIsInstance(self.logger, logging.Logger)
        self.assertEqual(self.logger.prefix, '/tmp/log/ioflo')
        self.assertTrue(self.logger.runner)  # runner generator is made when logger created

        log = logging.Log(name='test',
                          store=self.store,
                          kind='text',
                          baseFileName='',
                          rule=globaling.ONCE)

        self.assertEqual(log.rule, globaling.ONCE)
        self.assertEqual(log.action, log.once)

        heading = self.store.create('pose.heading').create(value = 0.0)
        log.addLoggee(tag = 'heading', loggee = 'pose.heading')
        self.logger.addLog(log)
        self.logger.resolve()  # resolves logs as well

        self.house.store.changeStamp(0.0)
        self.assertIs(log.stamp, None)
        status = self.logger.runner.send(globaling.START)  # reopens prepares and logs once
        self.assertEqual(log.formats, odict([('_time', '%s'), ('heading', odict([('value', '\t%s')]))]))

        self.store.advanceStamp(0.125)
        status = self.logger.runner.send(globaling.RUN)  # no log since not updated
        self.store.advanceStamp(0.125)
        self.assertEqual(self.store.stamp, 0.25)
        heading.value += 0.0  # updated but same value
        status = self.logger.runner.send(globaling.RUN)  # logs since updated
        self.store.advanceStamp(0.125)
        status = self.logger.runner.send(globaling.RUN)  # no log since not updated
        self.store.advanceStamp(0.125)
        self.assertEqual(self.store.stamp, 0.5)
        heading.value += 5.0  # update with different value
        status = self.logger.runner.send(globaling.RUN)  # logs since updated
        self.store.advanceStamp(0.125)
        status = self.logger.runner.send(globaling.STOP)  # not log since not updated

        log.reopen()
        log.file.seek(0)  # reopen appends so seek back to start
        lines = log.file.readlines()
        self.assertEqual(lines, ['text\tOnce\ttest\n', '_time\theading\n', '0.0\t0.0\n'])
        log.file.close()

        try:
            os.remove(log.path)  # remove to clean up
        except OSError as ex:
            pass


    def testLogNever(self):
        """
        Test log with never rule
        """
        console.terse("{0}\n".format(self.testLogNever.__doc__))

        self.assertEqual(self.house.store, self.store)
        self.assertIsInstance(self.logger, logging.Logger)
        self.assertEqual(self.logger.prefix, '/tmp/log/ioflo')
        self.assertTrue(self.logger.runner)  # runner generator is made when logger created

        log = logging.Log(name='test',
                          store=self.store,
                          kind='text',
                          baseFileName='',
                          rule=globaling.NEVER)

        self.assertEqual(log.rule, globaling.NEVER)
        self.assertEqual(log.action, log.never)

        heading = self.store.create('pose.heading').create(value = 0.0)
        log.addLoggee(tag = 'heading', loggee = 'pose.heading')
        self.logger.addLog(log)
        self.logger.resolve()  # resolves logs as well

        self.house.store.changeStamp(0.0)
        self.assertIs(log.stamp, None)
        status = self.logger.runner.send(globaling.START)  # reopens prepares and logs once
        self.assertEqual(log.formats, odict([('_time', '%s'), ('heading', odict([('value', '\t%s')]))]))

        self.store.advanceStamp(0.125)
        status = self.logger.runner.send(globaling.RUN)  # no log since not updated
        self.store.advanceStamp(0.125)
        self.assertEqual(self.store.stamp, 0.25)
        heading.value += 0.0  # updated but same value
        status = self.logger.runner.send(globaling.RUN)  # logs since updated
        self.store.advanceStamp(0.125)
        status = self.logger.runner.send(globaling.RUN)  # no log since not updated
        self.store.advanceStamp(0.125)
        self.assertEqual(self.store.stamp, 0.5)
        heading.value += 5.0  # update with different value
        status = self.logger.runner.send(globaling.RUN)  # logs since updated
        self.store.advanceStamp(0.125)
        status = self.logger.runner.send(globaling.STOP)  # not log since not updated

        log.reopen()
        log.file.seek(0)  # reopen appends so seek back to start
        lines = log.file.readlines()
        self.assertEqual(lines, ['text\tNever\ttest\n', '_time\theading\n'])
        log.file.close()

        try:
            os.remove(log.path)  # remove to clean up
        except OSError as ex:
            pass



    def testLogUpdate(self):
        """
        Test log with update rule
        """
        console.terse("{0}\n".format(self.testLogUpdate.__doc__))

        self.assertEqual(self.house.store, self.store)
        self.assertIsInstance(self.logger, logging.Logger)
        self.assertEqual(self.logger.prefix, '/tmp/log/ioflo')
        self.assertTrue(self.logger.runner)  # runner generator is made when logger created

        log = logging.Log(name='test',
                          store=self.store,
                          kind='text',
                          baseFileName='',
                          rule=globaling.UPDATE)

        self.assertEqual(log.rule, globaling.UPDATE)
        self.assertEqual(log.action, log.update)

        heading = self.store.create('pose.heading').create(value = 0.0)
        log.addLoggee(tag = 'heading', loggee = 'pose.heading')
        self.logger.addLog(log)
        self.logger.resolve()  # resolves logs as well

        self.house.store.changeStamp(0.0)
        self.assertIs(log.stamp, None)
        status = self.logger.runner.send(globaling.START)  # reopens prepares and logs once
        self.assertEqual(log.formats, odict([('_time', '%s'), ('heading', odict([('value', '\t%s')]))]))

        self.store.advanceStamp(0.125)
        status = self.logger.runner.send(globaling.RUN)  # no log since not updated
        self.store.advanceStamp(0.125)
        self.assertEqual(self.store.stamp, 0.25)
        heading.value += 0.0  # updated but same value
        status = self.logger.runner.send(globaling.RUN)  # logs since updated
        self.store.advanceStamp(0.125)
        status = self.logger.runner.send(globaling.RUN)  # no log since not updated
        self.store.advanceStamp(0.125)
        self.assertEqual(self.store.stamp, 0.5)
        heading.value += 5.0  # update with different value
        status = self.logger.runner.send(globaling.RUN)  # logs since updated
        self.store.advanceStamp(0.125)
        status = self.logger.runner.send(globaling.STOP)  # not log since not updated

        log.reopen()
        log.file.seek(0)  # reopen appends so seek back to start
        lines = log.file.readlines()
        self.assertEqual(lines, ['text\tUpdate\ttest\n',
                                '_time\theading\n',
                                '0.0\t0.0\n',
                                '0.25\t0.0\n',
                                '0.5\t5.0\n'])
        log.file.close()

        try:
            os.remove(log.path)  # remove to clean up
        except OSError as ex:
            pass



    def testLogUpdateFields(self):
        """
        Test log with update rule and passed in fields list
        """
        console.terse("{0}\n".format(self.testLogUpdateFields.__doc__))

        self.assertEqual(self.house.store, self.store)
        self.assertIsInstance(self.logger, logging.Logger)
        self.assertEqual(self.logger.prefix, '/tmp/log/ioflo')
        self.assertTrue(self.logger.runner)  # runner generator is made when logger created

        log = logging.Log(name='test',
                          store=self.store,
                          kind='text',
                          baseFileName='',
                          rule=globaling.UPDATE)

        self.assertEqual(log.rule, globaling.UPDATE)
        self.assertEqual(log.action, log.update)

        ned = self.store.create('pose.ned').create(north = 0.0, east=0.0, down=0.0)
        log.addLoggee(tag = 'ned', loggee = 'pose.ned', fields=['north', 'east'])
        self.logger.addLog(log)
        self.logger.resolve()  # resolves logs as well

        self.house.store.changeStamp(0.0)
        self.assertIs(log.stamp, None)
        status = self.logger.runner.send(globaling.START)  # reopens prepares and logs once
        self.assertEqual(log.formats, {'_time': '%s',
                                        'ned': odict([('north', '\t%s'), ('east', '\t%s')])}
                                       )

        self.store.advanceStamp(0.125)
        status = self.logger.runner.send(globaling.RUN)  # no log since not updated
        self.store.advanceStamp(0.125)
        self.assertEqual(self.store.stamp, 0.25)
        ned.update(north=0.0)  # updated but same value
        status = self.logger.runner.send(globaling.RUN)  # log since updated
        self.store.advanceStamp(0.125)
        ned.update(down=0.0)  # updated a field but not loggee field
        status = self.logger.runner.send(globaling.RUN)  # log since updated any field
        self.store.advanceStamp(0.125)
        self.assertEqual(self.store.stamp, 0.5)
        ned.update(north=5.0, east=7.0)  # update with different values
        status = self.logger.runner.send(globaling.RUN)  # log since updated
        self.store.advanceStamp(0.125)
        status = self.logger.runner.send(globaling.STOP)  # not log since not updated

        log.reopen()
        log.file.seek(0)  # reopen appends so seek back to start
        lines = log.file.readlines()
        self.assertEqual(lines, ['text\tUpdate\ttest\n',
                                '_time\tned.north\tned.east\n',
                                '0.0\t0.0\t0.0\n',
                                '0.25\t0.0\t0.0\n',
                                '0.375\t0.0\t0.0\n',
                                '0.5\t5.0\t7.0\n'])
        log.file.close()

        try:
            os.remove(log.path)  # remove to clean up
        except OSError as ex:
            pass


    def testLogChange(self):
        """
        Test log with change rule
        """
        console.terse("{0}\n".format(self.testLogChange.__doc__))

        self.assertEqual(self.house.store, self.store)
        self.assertIsInstance(self.logger, logging.Logger)
        self.assertEqual(self.logger.prefix, '/tmp/log/ioflo')
        self.assertTrue(self.logger.runner)  # runner generator is made when logger created

        log = logging.Log(name='test',
                          store=self.store,
                          kind='text',
                          baseFileName='',
                          rule=globaling.CHANGE)

        self.assertEqual(log.rule, globaling.CHANGE)
        self.assertEqual(log.action, log.change)

        heading = self.store.create('pose.heading').create(value = 0.0)
        log.addLoggee(tag = 'heading', loggee = 'pose.heading')
        self.logger.addLog(log)
        self.logger.resolve()  # resolves logs as well

        self.house.store.changeStamp(0.0)
        self.assertIs(log.stamp, None)
        status = self.logger.runner.send(globaling.START)  # reopens prepares and logs once
        self.assertEqual(log.formats, odict([('_time', '%s'), ('heading', odict([('value', '\t%s')]))]))
        self.assertTrue('heading' in log.lasts)

        self.store.advanceStamp(0.125)
        status = self.logger.runner.send(globaling.RUN)  # no log since not updated
        self.store.advanceStamp(0.125)
        self.assertEqual(self.store.stamp, 0.25)
        heading.value += 0.0  # updated but same value
        status = self.logger.runner.send(globaling.RUN)  # no log since updated but not changed
        self.store.advanceStamp(0.125)
        status = self.logger.runner.send(globaling.RUN)  # no log since not updated
        self.store.advanceStamp(0.125)
        self.assertEqual(self.store.stamp, 0.5)
        heading.value += 5.0  # update with different value
        status = self.logger.runner.send(globaling.RUN)  # logs since changed
        self.store.advanceStamp(0.125)
        status = self.logger.runner.send(globaling.STOP)  # not log since not updated

        log.reopen()
        log.file.seek(0)  # reopen appends so seek back to start
        lines = log.file.readlines()
        self.assertEqual(lines, ['text\tChange\ttest\n', '_time\theading\n', '0.0\t0.0\n', '0.5\t5.0\n'])
        log.file.close()

        try:
            os.remove(log.path)  # remove to clean up
        except OSError as ex:
            pass


    def testLogChangeFields(self):
        """
        Test log with update rule and passed in fields list
        """
        console.terse("{0}\n".format(self.testLogChangeFields.__doc__))

        self.assertEqual(self.house.store, self.store)
        self.assertIsInstance(self.logger, logging.Logger)
        self.assertEqual(self.logger.prefix, '/tmp/log/ioflo')
        self.assertTrue(self.logger.runner)  # runner generator is made when logger created

        log = logging.Log(name='test',
                          store=self.store,
                          kind='text',
                          baseFileName='',
                          rule=globaling.CHANGE)

        self.assertEqual(log.rule, globaling.CHANGE)
        self.assertEqual(log.action, log.change)

        ned = self.store.create('pose.ned').create(north = 0.0, east=0.0, down=0.0)
        fields = ['north', 'east']
        log.addLoggee(tag = 'ned', loggee = 'pose.ned', fields=fields)
        self.logger.addLog(log)
        self.logger.resolve()  # resolves logs as well

        self.house.store.changeStamp(0.0)
        self.assertIs(log.stamp, None)
        status = self.logger.runner.send(globaling.START)  # reopens prepares and logs once
        self.assertEqual(log.formats, {'_time': '%s',
                                    'ned': odict([('north', '\t%s'), ('east', '\t%s')])})
        self.assertTrue('ned' in log.lasts)
        last = log.lasts['ned']
        for field in fields:
            self.assertEqual(getattr(last, field), 0.0)

        self.store.advanceStamp(0.125)
        status = self.logger.runner.send(globaling.RUN)  # no log since not updated
        self.store.advanceStamp(0.125)
        self.assertEqual(self.store.stamp, 0.25)
        ned.update(north=0.0)  # updated but same value
        status = self.logger.runner.send(globaling.RUN)  # not log since not changed
        self.store.advanceStamp(0.125)
        ned.update(down=4.0)  # updated a field but not loggee field
        status = self.logger.runner.send(globaling.RUN)  # not log since not changed given field
        self.store.advanceStamp(0.125)
        self.assertEqual(self.store.stamp, 0.5)
        ned.update(north=5.0, east=7.0)  # update with different values
        status = self.logger.runner.send(globaling.RUN)  # log since changed given field
        self.store.advanceStamp(0.125)
        status = self.logger.runner.send(globaling.STOP)  # not log since not changed

        log.reopen()
        log.file.seek(0)  # reopen appends so seek back to start
        lines = log.file.readlines()
        self.assertEqual(lines, ['text\tChange\ttest\n',
                                '_time\tned.north\tned.east\n',
                                '0.0\t0.0\t0.0\n',
                                '0.5\t5.0\t7.0\n'])
        log.file.close()

        try:
            os.remove(log.path)  # remove to clean up
        except OSError as ex:
            pass


    def testLogStreak(self):
        """
        Test log with streak rule
        """
        console.terse("{0}\n".format(self.testLogStreak.__doc__))

        self.assertEqual(self.house.store, self.store)
        self.assertIsInstance(self.logger, logging.Logger)
        self.assertEqual(self.logger.prefix, '/tmp/log/ioflo')
        self.assertTrue(self.logger.runner)  # runner generator is made when logger created

        log = logging.Log(name='test',
                          store=self.store,
                          kind='text',
                          baseFileName='',
                          rule=globaling.STREAK)

        self.assertEqual(log.rule, globaling.STREAK)
        self.assertEqual(log.action, log.streak)

        heading = self.store.create('pose.heading').create(value = 0.0)
        log.addLoggee(tag = 'heading', loggee = 'pose.heading')
        self.logger.addLog(log)
        self.logger.resolve()  # resolves logs as well

        self.house.store.changeStamp(0.0)
        self.assertIs(log.stamp, None)
        status = self.logger.runner.send(globaling.START)  # reopens prepares and logs once
        self.assertEqual(log.formats, odict([('_time', '%s'), ('heading', odict([('value', '\t%s')]))]))

        self.store.advanceStamp(0.125)
        status = self.logger.runner.send(globaling.RUN)  # no log since not updated
        self.store.advanceStamp(0.125)
        self.assertEqual(self.store.stamp, 0.25)
        heading.value += 0.0  # updated but same value
        status = self.logger.runner.send(globaling.RUN)  # no log since updated but not changed
        self.store.advanceStamp(0.125)
        status = self.logger.runner.send(globaling.RUN)  # no log since not updated
        self.store.advanceStamp(0.125)
        self.assertEqual(self.store.stamp, 0.5)
        heading.value += 5.0  # update with different value
        status = self.logger.runner.send(globaling.RUN)  # logs since changed
        self.store.advanceStamp(0.125)
        status = self.logger.runner.send(globaling.STOP)  # not log since not updated

        log.reopen()
        log.file.seek(0)  # reopen appends so seek back to start
        lines = log.file.readlines()
        self.assertEqual(lines, ['text\tStreak\ttest\n',
                                '_time\theading\n',
                                '0.0\t0.0\n',
                                '0.125\t0.0\n',
                                '0.25\t0.0\n',
                                '0.375\t0.0\n',
                                '0.5\t5.0\n',
                                '0.625\t5.0\n'])
        log.file.close()

        heading.value = ["hello", "how", "are", "you", 5.0, 6, 7]

        self.assertEqual(log.stamp, 0.625)
        status = self.logger.runner.send(globaling.START)  # reopens prepares and logs once
        self.assertEqual(log.formats, odict([('_time', '%s'), ('heading', odict([('value', '\t%s')]))]))
        #self.assertTrue('heading' in log.lasts)

        self.store.advanceStamp(0.125)
        status = self.logger.runner.send(globaling.RUN)  # no log since not updated
        self.store.advanceStamp(0.125)
        heading.value.append(10.0)
        status = self.logger.runner.send(globaling.RUN)  # no log since updated but not changed
        self.store.advanceStamp(0.125)
        status = self.logger.runner.send(globaling.RUN)  # no log since not updated
        self.store.advanceStamp(0.125)
        heading.value.append(15)
        heading.value.append(20)
        status = self.logger.runner.send(globaling.RUN)  # logs since changed
        self.store.advanceStamp(0.125)
        status = self.logger.runner.send(globaling.STOP)  # not log since not updated

        log.reopen()
        log.file.seek(0)  # reopen appends so seek back to start
        lines = log.file.readlines()
        self.assertEqual(lines, ['text\tStreak\ttest\n',
                                '_time\theading\n',
                                '0.0\t0.0\n',
                                '0.125\t0.0\n',
                                '0.25\t0.0\n',
                                '0.375\t0.0\n',
                                '0.5\t5.0\n',
                                '0.625\t5.0\n',
                                '0.625\thello\n',
                                '0.625\thow\n',
                                '0.625\tare\n',
                                '0.625\tyou\n',
                                '0.625\t5.0\n',
                                '0.625\t6\n',
                                '0.625\t7\n',
                                '0.875\t10.0\n',
                                '1.125\t15\n',
                                '1.125\t20\n'])
        log.file.close()

        try:
            os.remove(log.path)  # remove to clean up
        except OSError as ex:
            pass


    def testLogStreakFields(self):
        """
        Test log with streak rule and fields
        """
        console.terse("{0}\n".format(self.testLogStreakFields.__doc__))

        self.assertEqual(self.house.store, self.store)
        self.assertIsInstance(self.logger, logging.Logger)
        self.assertEqual(self.logger.prefix, '/tmp/log/ioflo')
        self.assertTrue(self.logger.runner)  # runner generator is made when logger created

        log = logging.Log(name='test',
                          store=self.store,
                          kind='text',
                          baseFileName='',
                          rule=globaling.STREAK)

        self.assertEqual(log.rule, globaling.STREAK)
        self.assertEqual(log.action, log.streak)

        stuff = self.store.create('big.stuff').create([('sack', []),
                                                   ('bag', [])])
        fields = ['bag', 'dummy']  # reorder to second field gets logged
        log.addLoggee(tag = 'puff', loggee = 'big.stuff', fields=fields)
        self.logger.addLog(log)
        self.logger.resolve()  # resolves logs as well

        self.house.store.changeStamp(0.0)
        self.assertIs(log.stamp, None)
        status = self.logger.runner.send(globaling.START)  # reopens prepares and logs once
        self.assertEqual(log.formats, odict([('_time', '%s'), ('puff', odict([('bag', '\t%s')]))]))

        self.store.advanceStamp(0.125)
        status = self.logger.runner.send(globaling.RUN)  # no log since empty
        self.store.advanceStamp(0.125)
        self.assertEqual(self.store.stamp, 0.25)
        stuff.data.bag.extend(["Rain", "Falls"])
        stuff.data.sack.extend(["Sun", "Shines"])
        status = self.logger.runner.send(globaling.RUN)  # log since not empty
        self.store.advanceStamp(0.125)
        status = self.logger.runner.send(globaling.RUN)  # no log since empty
        self.store.advanceStamp(0.125)
        self.assertEqual(self.store.stamp, 0.5)
        stuff.data.bag.extend(["Sky", "Blue", "Tree", "Green"])
        status = self.logger.runner.send(globaling.RUN)  # logs since not empty
        self.store.advanceStamp(0.125)
        status = self.logger.runner.send(globaling.STOP)  # not log since not updated

        log.reopen()
        log.file.seek(0)  # reopen appends so seek back to start
        lines = log.file.readlines()
        self.assertEqual(lines, ['text\tStreak\ttest\n',
                                '_time\tpuff\n',
                                '0.25\tRain\n',
                                '0.25\tFalls\n',
                                '0.5\tSky\n',
                                '0.5\tBlue\n',
                                '0.5\tTree\n',
                                '0.5\tGreen\n'])
        log.close()

        try:
            os.remove(log.path)  # remove to clean up
        except OSError as ex:
            pass


    def testLogDeck(self):
        """
        Test log with deck rule
        """
        console.terse("{0}\n".format(self.testLogDeck.__doc__))

        self.assertEqual(self.house.store, self.store)
        self.assertIsInstance(self.logger, logging.Logger)
        self.assertEqual(self.logger.prefix, '/tmp/log/ioflo')
        self.assertTrue(self.logger.runner)  # runner generator is made when logger created

        log = logging.Log(name='test',
                          store=self.store,
                          kind='text',
                          baseFileName='',
                          rule=globaling.DECK)

        self.assertEqual(log.rule, globaling.DECK)
        self.assertEqual(log.action, log.deck)

        ned = self.store.create('pose.ned')
        fields = ['north', 'east']
        log.addLoggee(tag = 'ned', loggee = 'pose.ned', fields=fields)
        self.logger.addLog(log)
        self.logger.resolve()  # resolves logs as well

        ned.push(odict(north=0.0, east=0.0, down=0.0))
        ned.push(odict(north=5.0, east=4.0, down=3.0))
        ned.push(odict(north=6.0, east=3.0, down=2.0))
        ned.push(odict(east=2.0, down=1.0))
        ned.push(odict(down=0.0))
        ned.push(odict(north=7.0, east=4.0, down=3.0))
        ned.push(["hi", "there"])
        ned.push(odict(north=8.0, east=5.0, down=4.0))

        self.assertEqual(ned.deck, Deck([odict([('down', 0.0),('north', 0.0), ('east', 0.0)]),
                                         odict([('down', 3.0), ('north', 5.0), ('east', 4.0)]),
                                         odict([('down', 2.0), ('north', 6.0), ('east', 3.0)]),
                                         odict([('down', 1.0), ('east', 2.0)]),
                                         odict([('down', 0.0)]),
                                         odict([('down', 3.0), ('north', 7.0), ('east', 4.0)]),
                                         ['hi', 'there'],
                                         odict([('down', 4.0), ('north', 8.0), ('east', 5.0)])]))

        self.house.store.changeStamp(0.0)
        self.assertIs(log.stamp, None)
        status = self.logger.runner.send(globaling.START)  # reopens prepares and logs once
        self.assertEqual(log.formats, {'_time': '%s',
                                       'ned': odict([('north', '\t%s'), ('east', '\t%s')])})

        self.store.advanceStamp(0.125)
        status = self.logger.runner.send(globaling.RUN)  # no log since not updated
        self.store.advanceStamp(0.125)
        self.assertEqual(self.store.stamp, 0.25)
        ned.push(odict(north=9.0, east=6.0, down=5.0))  # push another
        status = self.logger.runner.send(globaling.RUN)  # not log since not changed
        self.store.advanceStamp(0.125)
        status = self.logger.runner.send(globaling.RUN)  # not log since not changed given field
        self.store.advanceStamp(0.125)
        self.assertEqual(self.store.stamp, 0.5)
        ned.push(odict(north=10.0, east=7.0, down=6.0))  # push
        status = self.logger.runner.send(globaling.RUN)  # log since changed given field
        self.store.advanceStamp(0.125)
        status = self.logger.runner.send(globaling.STOP)  # not log since not changed

        log.reopen()
        log.file.seek(0)  # reopen appends so seek back to start
        lines = log.file.readlines()
        self.assertEqual(lines, ['text\tDeck\ttest\n',
                                '_time\tned.north\tned.east\n',
                                '0.0\t0.0\t0.0\n',
                                '0.0\t5.0\t4.0\n',
                                '0.0\t6.0\t3.0\n',
                                '0.0\t\t2.0\n',
                                '0.0\t\t\n',
                                '0.0\t7.0\t4.0\n',
                                '0.0\t8.0\t5.0\n',
                                '0.25\t9.0\t6.0\n',
                                '0.5\t10.0\t7.0\n'])
        log.file.close()

        try:
            os.remove(log.path)  # remove to clean up
        except OSError as ex:
            pass


class HouseTestCase(testing.HouseIofloTestCase):
    """
    Example TestCase
    """

    def setUp(self):
        """
        Call super if override so House Framer and Frame are setup correctly
        """
        super(HouseTestCase, self).setUp()

    def tearDown(self):
        """
        Call super if override so House Framer and Frame are torn down correctly
        """
        super(HouseTestCase, self).tearDown()


    def testSetup(self):
        """
        Test creating setting up a logger
        """
        console.terse("{0}\n".format(self.testSetup.__doc__))
        self.assertEqual(self.house.store, self.store)

        prefix = "/tmp/log/ioflo"
        flush = 2.0
        cycle = 0.0
        keep =  0
        logger = logging.Logger(name="LoggerTest",
                                store=self.store,
                                schedule=globaling.ACTIVE,
                                prefix=prefix,
                                flushPeriod=flush,
                                cycle=cycle,
                                keep=keep)

        self.assertEqual(logger.flushPeriod, flush)
        self.assertEqual(logger.prefix, '/tmp/log/ioflo')
        self.assertTrue(logger.runner)  # runner generator is made when logger created
        self.assertEqual(logger.cyclePeriod, cycle)
        self.assertEqual(logger.keep, keep)
        self.assertEqual(logger.flushStamp, 0.0)
        self.assertEqual(logger.path, '')

        self.assertEqual(logger.logs, [])

        self.house.taskers.append(logger)
        self.house.mids.append(logger)
        self.house.orderTaskables()

        self.house.store.changeStamp(0.0)
        self.assertEqual(logger.stamp, 0.0)

        status = logger.runner.send(globaling.START)  # reopens prepares
        self.assertTrue(logger.path.startswith("/tmp/log/ioflo/HouseTest/LoggerTest_"))


    def testCycle(self):
        """
        Test creating setting up a logger with log to rotate
        """
        console.terse("{0}\n".format(self.testCycle.__doc__))
        self.assertEqual(self.house.store, self.store)

        prefix = "/tmp/log/ioflo"
        flush = 3.0
        cyclePeriod = 0.5
        keep =  2
        fileSize = 10
        logger = logging.Logger(name="LoggerTest",
                                store=self.store,
                                schedule=globaling.ACTIVE,
                                prefix=prefix,
                                flushPeriod=flush,
                                keep=keep,
                                cyclePeriod=cyclePeriod,
                                fileSize=fileSize)

        self.assertEqual(logger.flushPeriod, flush)
        self.assertEqual(logger.prefix, '/tmp/log/ioflo')
        self.assertTrue(logger.runner)  # runner generator is made when logger created
        self.assertEqual(logger.cyclePeriod, cyclePeriod)
        self.assertEqual(logger.keep, keep)
        self.assertEqual(logger.fileSize, fileSize)
        self.assertEqual(logger.cycleStamp, 0.0)
        self.assertEqual(logger.flushStamp, 0.0)
        self.assertEqual(logger.path, '')

        self.assertEqual(logger.logs, [])

        self.house.taskers.append(logger)
        self.house.mids.append(logger)
        self.house.orderTaskables()

        self.house.store.changeStamp(0.0)
        self.assertEqual(logger.stamp, 0.0)

        log = logging.Log(name='test',
                          store=self.store,
                          kind='text',
                          baseFileName='',
                          rule=globaling.ALWAYS)

        self.assertEqual(log.baseFilename, log.name)
        self.assertEqual(log.path, '')
        self.assertEqual(log.file, None)
        self.assertEqual(log.kind, 'text')

        self.assertEqual(log.rule, globaling.ALWAYS)
        self.assertEqual(log.action, log.always)

        logger.addLog(log)

        heading = self.store.create('pose.heading').create(value = 0.0)
        #position = self.store.create('pose.position').create([("north", 10.0), ("east", 5.0)])

        log.addLoggee(tag = 'heading', loggee = 'pose.heading')
        #log.addLoggee(tag = 'pos', loggee = 'pose.position')

        logger.resolve()  # resolves logs as well

        self.house.store.changeStamp(0.0)
        self.assertIs(log.stamp, None)

        status = logger.runner.send(globaling.START)  # reopens prepares and logs once

        self.assertTrue(logger.path.startswith('/tmp/log/ioflo/HouseTest/LoggerTest_'))
        self.assertTrue(log.path.startswith(logger.path))
        self.assertTrue(log.path.endswith(log.baseFilename + '.txt'))
        self.assertTrue(log.file)

        self.assertEqual(len(log.paths), keep+1)
        for k, path in enumerate(log.paths[1:]):
            self.assertTrue(path.startswith(logger.path))
            base = os.path.basename(path)
            root, ext = os.path.splitext(base)
            self.assertTrue(root.endswith("{0:02}".format(k+1)))

        for i in range(16):
            self.store.advanceStamp(0.125)
            heading.value += 1.0
            status = logger.runner.send(globaling.RUN)

        self.store.advanceStamp(0.125)
        heading.value += 1.0
        status = logger.runner.send(globaling.STOP)  # logs once and closes logs

        path0, path1, path2 = log.paths

        file0 = open(path0, "r")
        lines = file0.readlines()
        self.assertEqual(lines, ['text\tAlways\ttest\n', '_time\theading\n',
                                 '2.125\t17.0\n'])
        file0.close()

        file1 = open(path1, "r")
        lines = file1.readlines()
        self.assertEqual(lines, ['text\tAlways\ttest\n',
                                '_time\theading\n',
                                '1.625\t13.0\n',
                                '1.75\t14.0\n',
                                '1.875\t15.0\n',
                                '2.0\t16.0\n'])
        file1.close()

        file2 = open(path2, "r")
        lines = file2.readlines()
        self.assertEqual(lines, ['text\tAlways\ttest\n',
                                '_time\theading\n',
                                '1.125\t9.0\n',
                                '1.25\t10.0\n',
                                '1.375\t11.0\n',
                                '1.5\t12.0\n'])
        file2.close()

        for path in log.paths:  # remove files to clean up
            try:
                os.remove(path)
            except OSError as ex:
                pass


    def testCycleReuse(self):
        """
        Test logger with rotate and non unique directory name
        """
        console.terse("{0}\n".format(self.testCycleReuse.__doc__))
        self.assertEqual(self.house.store, self.store)

        prefix = "/tmp/log/ioflo"
        flush = 3.0
        cyclePeriod = 0.5
        keep =  2
        fileSize = 50
        logger = logging.Logger(name="LoggerTest",
                                store=self.store,
                                schedule=globaling.ACTIVE,
                                prefix=prefix,
                                flushPeriod=flush,
                                keep=keep,
                                cyclePeriod=cyclePeriod,
                                fileSize=fileSize,
                                reuse=True)

        self.assertEqual(logger.flushPeriod, flush)
        self.assertEqual(logger.prefix, '/tmp/log/ioflo')
        self.assertTrue(logger.runner)  # runner generator is made when logger created
        self.assertEqual(logger.cyclePeriod, cyclePeriod)
        self.assertEqual(logger.keep, keep)
        self.assertEqual(logger.fileSize, fileSize)
        self.assertEqual(logger.cycleStamp, 0.0)
        self.assertEqual(logger.flushStamp, 0.0)
        self.assertEqual(logger.path, '')
        self.assertTrue(logger.reuse)

        self.assertEqual(logger.logs, [])

        self.house.taskers.append(logger)
        self.house.mids.append(logger)
        self.house.orderTaskables()

        self.house.store.changeStamp(0.0)
        self.assertEqual(logger.stamp, 0.0)

        log = logging.Log(name='test',
                          store=self.store,
                          kind='text',
                          baseFileName='',
                          rule=globaling.ALWAYS)

        self.assertEqual(log.baseFilename, log.name)
        self.assertEqual(log.path, '')
        self.assertEqual(log.file, None)
        self.assertEqual(log.kind, 'text')

        self.assertEqual(log.rule, globaling.ALWAYS)
        self.assertEqual(log.action, log.always)

        logger.addLog(log)

        heading = self.store.create('pose.heading').create(value = 0.0)
        #position = self.store.create('pose.position').create([("north", 10.0), ("east", 5.0)])

        log.addLoggee(tag = 'heading', loggee = 'pose.heading')
        #log.addLoggee(tag = 'pos', loggee = 'pose.position')

        logger.resolve()  # resolves logs as well

        self.house.store.changeStamp(0.0)
        self.assertIs(log.stamp, None)

        status = logger.runner.send(globaling.START)  # reopens prepares and logs once

        self.assertTrue(logger.path.startswith('/tmp/log/ioflo/HouseTest/LoggerTest'))
        self.assertTrue(log.path.startswith(logger.path))
        self.assertTrue(log.path.endswith(log.baseFilename + '.txt'))
        self.assertTrue(log.file)

        self.assertEqual(len(log.paths), keep+1)
        for k, path in enumerate(log.paths[1:]):
            self.assertTrue(path.startswith(logger.path))
            base = os.path.basename(path)
            root, ext = os.path.splitext(base)
            self.assertTrue(root.endswith("{0:02}".format(k+1)))

        for i in range(16):
            self.store.advanceStamp(0.125)
            heading.value += 1.0
            status = logger.runner.send(globaling.RUN)

        self.store.advanceStamp(0.125)
        heading.value += 1.0
        status = logger.runner.send(globaling.STOP)  # logs once and closes logs

        path0, path1, path2 = log.paths

        file0 = open(path0, "r")
        lines = file0.readlines()
        self.assertEqual(lines, ['text\tAlways\ttest\n',
                                 '_time\theading\n',
                                 '2.125\t17.0\n'])
        file0.close()

        file1 = open(path1, "r")
        lines = file1.readlines()
        self.assertEqual(lines,['text\tAlways\ttest\n',
                                '_time\theading\n',
                                '1.625\t13.0\n',
                                '1.75\t14.0\n',
                                '1.875\t15.0\n',
                                '2.0\t16.0\n'] )
        file1.close()

        file2 = open(path2, "r")
        lines = file2.readlines()
        self.assertEqual(lines, ['text\tAlways\ttest\n',
                                '_time\theading\n',
                                '1.125\t9.0\n',
                                '1.25\t10.0\n',
                                '1.375\t11.0\n',
                                '1.5\t12.0\n'])
        file2.close()

        # restart for reuse
        status = logger.runner.send(globaling.START)  # reopens prepares and logs once

        self.assertTrue(logger.path.startswith('/tmp/log/ioflo/HouseTest/LoggerTest'))
        self.assertTrue(log.path.startswith(logger.path))
        self.assertTrue(log.path.endswith(log.baseFilename + '.txt'))
        self.assertTrue(log.file)

        self.assertEqual(len(log.paths), keep+1)
        for k, path in enumerate(log.paths[1:]):
            self.assertTrue(path.startswith(logger.path))
            base = os.path.basename(path)
            root, ext = os.path.splitext(base)
            self.assertTrue(root.endswith("{0:02}".format(k+1)))

        for i in range(3):
            self.store.advanceStamp(0.125)
            heading.value += 1.0
            status = logger.runner.send(globaling.RUN)

        self.store.advanceStamp(0.125)
        heading.value += 1.0
        status = logger.runner.send(globaling.STOP)  # logs once and closes logs

        path0, path1, path2 = log.paths

        file0 = open(path0, "r")
        lines = file0.readlines()
        self.assertEqual(lines, ['text\tAlways\ttest\n',
                                 '_time\theading\n',
                                 '2.625\t21.0\n'])
        file0.close()

        file1 = open(path1, "r")
        lines = file1.readlines()
        self.assertEqual(lines, ['text\tAlways\ttest\n',
                                '_time\theading\n',
                                '2.125\t17.0\n',
                                '2.125\t17.0\n',
                                '2.25\t18.0\n',
                                '2.375\t19.0\n',
                                '2.5\t20.0\n'])
        file1.close()

        file2 = open(path2, "r")
        lines = file2.readlines()
        self.assertEqual(lines, ['text\tAlways\ttest\n',
                                '_time\theading\n',
                                '1.625\t13.0\n',
                                '1.75\t14.0\n',
                                '1.875\t15.0\n',
                                '2.0\t16.0\n'])
        file2.close()

        for path in log.paths:  # remove files to clean up
            try:
                os.remove(path)
            except OSError as ex:
                pass

    def testReuse(self):
        """
        Test logger with non unique directory name
        """
        console.terse("{0}\n".format(self.testReuse.__doc__))
        self.assertEqual(self.house.store, self.store)

        prefix = "/tmp/log/ioflo"
        flush = 3.0
        logger = logging.Logger(name="LoggerTest",
                                store=self.store,
                                schedule=globaling.ACTIVE,
                                prefix=prefix,
                                flushPeriod=flush,
                                reuse=True)

        self.assertEqual(logger.flushPeriod, flush)
        self.assertEqual(logger.prefix, '/tmp/log/ioflo')
        self.assertTrue(logger.runner)  # runner generator is made when logger created
        self.assertEqual(logger.cyclePeriod, 0)
        self.assertEqual(logger.keep, 0)
        self.assertEqual(logger.fileSize, 0)
        self.assertEqual(logger.cycleStamp, 0.0)
        self.assertEqual(logger.flushStamp, 0.0)
        self.assertEqual(logger.path, '')
        self.assertTrue(logger.reuse)

        self.assertEqual(logger.logs, [])

        self.house.taskers.append(logger)
        self.house.mids.append(logger)
        self.house.orderTaskables()

        self.house.store.changeStamp(0.0)
        self.assertEqual(logger.stamp, 0.0)

        log = logging.Log(name='test',
                          store=self.store,
                          kind='text',
                          baseFileName='',
                          rule=globaling.ALWAYS)

        self.assertEqual(log.baseFilename, log.name)
        self.assertEqual(log.path, '')
        self.assertEqual(log.file, None)
        self.assertEqual(log.kind, 'text')

        self.assertEqual(log.rule, globaling.ALWAYS)
        self.assertEqual(log.action, log.always)

        logger.addLog(log)

        heading = self.store.create('pose.heading').create(value = 0.0)
        #position = self.store.create('pose.position').create([("north", 10.0), ("east", 5.0)])

        log.addLoggee(tag = 'heading', loggee = 'pose.heading')
        #log.addLoggee(tag = 'pos', loggee = 'pose.position')

        logger.resolve()  # resolves logs as well

        self.house.store.changeStamp(0.0)
        self.assertIs(log.stamp, None)

        status = logger.runner.send(globaling.START)  # reopens prepares and logs once

        self.assertTrue(logger.path.startswith('/tmp/log/ioflo/HouseTest/LoggerTest'))
        self.assertTrue(log.path.startswith(logger.path))
        self.assertTrue(log.path.endswith(log.baseFilename + '.txt'))
        self.assertTrue(log.file)

        self.assertEqual(len(log.paths), 0)

        for i in range(16):
            self.store.advanceStamp(0.125)
            heading.value += 1.0
            status = logger.runner.send(globaling.RUN)

        self.store.advanceStamp(0.125)
        heading.value += 1.0
        status = logger.runner.send(globaling.STOP)  # logs once and closes logs

        file = open(log.path, "r")
        lines = file.readlines()
        self.assertEqual(lines, ['text\tAlways\ttest\n',
                                '_time\theading\n',
                                '0.0\t0.0\n',
                                '0.125\t1.0\n',
                                '0.25\t2.0\n',
                                '0.375\t3.0\n',
                                '0.5\t4.0\n',
                                '0.625\t5.0\n',
                                '0.75\t6.0\n',
                                '0.875\t7.0\n',
                                '1.0\t8.0\n',
                                '1.125\t9.0\n',
                                '1.25\t10.0\n',
                                '1.375\t11.0\n',
                                '1.5\t12.0\n',
                                '1.625\t13.0\n',
                                '1.75\t14.0\n',
                                '1.875\t15.0\n',
                                '2.0\t16.0\n',
                                '2.125\t17.0\n'])
        file.close()

        # restart for reuse
        self.house.store.changeStamp(0.0)
        log.stamp = None
        self.assertIs(log.stamp, None)

        status = logger.runner.send(globaling.START)  # reopens prepares and logs once

        self.assertTrue(logger.path.startswith('/tmp/log/ioflo/HouseTest/LoggerTest'))
        self.assertTrue(log.path.startswith(logger.path))
        self.assertTrue(log.path.endswith(log.baseFilename + '.txt'))
        self.assertTrue(log.file)

        self.assertEqual(len(log.paths), 0)

        status = logger.runner.send(globaling.START)  # reopens prepares and logs once

        self.assertTrue(logger.path.startswith('/tmp/log/ioflo/HouseTest/LoggerTest'))
        self.assertTrue(log.path.startswith(logger.path))
        self.assertTrue(log.path.endswith(log.baseFilename + '.txt'))
        self.assertTrue(log.file)

        self.assertEqual(len(log.paths), 0)

        for i in range(3):
            self.store.advanceStamp(0.125)
            heading.value += 1.0
            status = logger.runner.send(globaling.RUN)

        self.store.advanceStamp(0.125)
        heading.value += 1.0
        status = logger.runner.send(globaling.STOP)  # logs once and closes logs

        file = open(log.path, "r")
        lines = file.readlines()
        self.assertEqual(lines, ['text\tAlways\ttest\n',
                                '_time\theading\n',
                                '0.0\t0.0\n',
                                '0.125\t1.0\n',
                                '0.25\t2.0\n',
                                '0.375\t3.0\n',
                                '0.5\t4.0\n',
                                '0.625\t5.0\n',
                                '0.75\t6.0\n',
                                '0.875\t7.0\n',
                                '1.0\t8.0\n',
                                '1.125\t9.0\n',
                                '1.25\t10.0\n',
                                '1.375\t11.0\n',
                                '1.5\t12.0\n',
                                '1.625\t13.0\n',
                                '1.75\t14.0\n',
                                '1.875\t15.0\n',
                                '2.0\t16.0\n',
                                '2.125\t17.0\n',
                                '0.0\t17.0\n',
                                '0.0\t17.0\n',
                                '0.125\t18.0\n',
                                '0.25\t19.0\n',
                                '0.375\t20.0\n',
                                '0.5\t21.0\n'])
        file.close()

        try:
            os.remove(log.path)  # remove to clean up
        except OSError as ex:
            pass


def runOneLogger(test):
    '''
    Unittest Runner
    '''
    test = LoggerTestCase(test)
    suite = unittest.TestSuite([test])
    unittest.TextTestRunner(verbosity=2).run(suite)

def runOneHouse(test):
    '''
    Unittest Runner
    '''
    test = HouseTestCase(test)
    suite = unittest.TestSuite([test])
    unittest.TextTestRunner(verbosity=2).run(suite)

def runSome():
    """ Unittest runner """
    tests =  []
    names = [
                'testLogger',
                'testLogAlways',
                'testLogOnce',
                'testLogNever',
                'testLogUpdate',
                'testLogUpdateFields',
                'testLogChange',
                'testLogChangeFields',
                'testLogStreak',
                'testLogStreakFields',
                'testLogDeck',
            ]
    tests.extend(map(LoggerTestCase, names))

    names = [
                'testSetup',
                'testCycle',
                'testCycleReuse',
                'testReuse',
            ]
    tests.extend(map(HouseTestCase, names))

    suite = unittest.TestSuite(tests)
    unittest.TextTestRunner(verbosity=2).run(suite)

def runAll():
    """ Unittest runner """
    suite = unittest.TestSuite()
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(LoggerTestCase))
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(HouseTestCase))
    unittest.TextTestRunner(verbosity=2).run(suite)

if __name__ == '__main__' and __package__ is None:

    #console.reinit(verbosity=console.Wordage.concise)

    #runAll() #run all unittests

    runSome()#only run some

    #runOneLogger('testLogStreak')
    #runOneHouse('testCycleReuse')



"""
Device Base Package

"""
from __future__ import absolute_import, division, print_function

import struct
from binascii import hexlify
from collections import deque, namedtuple
import enum
import socket

from ...aid.sixing import *
from ...aid.odicting import odict
from ...aid.byting import bytify, unbytify, packify, unpackify
from ...aid.eventing import eventify, tagify
from ...aid import getConsole
from .. import aioing
from .protoing import MixIn

console = getConsole()

class Device(MixIn):
    """
    Device Class
    """
    def __init__(self,
                 stack,
                 uid=None,
                 name=None,
                 ha=None,
                 kind=None,
                 ):
        """
        Initialization method for instance

        Parameters:
            stack is Stack managing this device required
            name is user friendly name of device
            uid  is unique device id
            ha is device host address
            kind is type of device

        Attributes:
            .stack is Stack managing this device required
            .name is user friendly name of device
            .uid  is unique device id per channel or site
            .ha is device host address
            .kind is type of device

        """
        self.stack = stack
        self.uid = uid if uid is not None else stack.nextUid()
        self.name = name if name is not None else "Device{0}".format(self.uid)
        self.ha = ha if ha is not None else ''
        self.kind = kind

    def show(self):
        """
        Display device data
        """
        result = ("Device: name={0} uid={1} ha={2} kind={3}\n".format(
                    self.name,
                    self.uid,
                    self.ha,
                    self.kind))
        return result

    def process(self):
        """
        Timer based processing
        """
        pass

    def receive(self, rx):
        """
        Process received rx msg/pkt/data.
        """
        pass

class LocalDevice(Device):
    """
    Local Device Class
    """
    def __init__(self,
                 stack,
                 **kwa):
        """
        Initialization method for instance

        Assumes local device in stack is created before any remotes are added

        Inherited Parameters:
            stack is Stack managing this device required
            name is user friendly name of device
            uid  is unique device id
            ha is device host address
            kind is type of device

        Inherited Attributes:
            .stack is Stack managing this device required
            .name is user friendly name of device
            .uid  is unique device id per channel or site
            .ha is device host address
            .kind is type of device

        """
        super(LocalDevice, self).__init__(stack=stack, **kwa)


class RemoteDevice(Device):
    """
    Remote Device Class
    """
    def __init__(self,
                 stack,
                 uid=None,
                 **kwa):
        """
        Initialization method for instance

        Inherited Parameters:
            stack is Stack managing this device required
            name is user friendly name of device
            uid  is unique device id
            ha is device host address
            kind is type of device

        Inherited Attributes:
            .stack is Stack managing this device required
            .name is user friendly name of device
            .uid  is unique device id per channel or site
            .ha is device host address
            .kind is type of device

        """
        if uid is None:
            uid = stack.nextUid()
            while uid in stack.remotes or uid in (stack.local.uid,):
                uid = stack.nextUid()

        super(RemoteDevice, self).__init__(stack=stack, uid=uid, **kwa)

    def receive(self, msg):
        """
        Process received rx msg/pkt/data.
        """
        if msg is not None:
            self.stack.rxMsgs.append(msg)


class SingleRemoteDevice(Device):
    """
    Remote Device Class when only one remote in stack .remote
    Affects how uid is assigned
    """
    def __init__(self,
                 stack,
                 uid=None,
                 uids=None,
                 **kwa):
        """
        Initialization method for instance

        Inherited Parameters:
            stack is Stack managing this device required
            name is user friendly name of device
            uid  is unique device id
            ha is device host address
            kind is type of device

        Parameters:
            uids is sequence or set of used uids to not use for remote if uid not provided

        Inherited Attributes:
            .stack is Stack managing this device required
            .name is user friendly name of device
            .uid  is unique device id per channel or site
            .ha is device host address
            .kind is type of device

        Attributes:
            .uids is sequence or set of used uids to not use for remote if uid not provided

        """
        if uid is None:
            uids = set(uids) if uids is not None else set()
            if hasattr(stack, 'local'):
                uids.add(stack.local.uid)
            uid = stack.nextUid()
            while uid in uids:
                uid = stack.nextUid()

        super(SingleRemoteDevice, self).__init__(stack=stack, uid=uid, **kwa)


class IpDevice(Device):
    """
    IP device
    """

    def __init__(self,
                 stack,
                 ha=None,
                 **kwa):
        """
        Initialization method for instance

        Inherited Parameters:
            stack is Stack managing this device required
            name is user friendly name of device
            uid  is unique device id
            ha is device  udp host address, a duple (host,port)
            kind is type of device

        Parameters:


        Inherited Attributes:
            .stack is Stack managing this device required
            .name is user friendly name of device
            .uid  is unique device id per channel or site
            .ha is device udp host address, a duple (host,port)
            .kind is type of device

        Attributes:

        """
        if ha:
            host, port = ha
            host = aioing.normalizeHost(host)
            if host in ('0.0.0.0',):
                host = '127.0.0.1'
            elif host in ("::", "0:0:0:0:0:0:0:0"):
                host = "::1"
            ha = (host, port)
        else:
            ha = ('127.0.0.1', stack.Port)

        super(IpDevice, self).__init__(stack=stack, ha=ha, **kwa)

    @property
    def host(self):
        """
        Property that returns host of local interface ha duple (host, port)
        """
        return self.ha[0]

    @host.setter
    def host(self, value):
        """
        Setter for host property
        """
        host, port = self.local.ha
        self.local.ha = (value, port)

    @property
    def port(self):
        """
        Property that returns host of local interface ha duple (host, port)
        """
        return self.ha[1]

    @port.setter
    def port(self, value):
        """
        Setter for host property
        """
        host, port = self.local.ha
        self.local.ha = (host, value)


class IpLocalDevice(IpDevice, LocalDevice):
    """
    Ip LocalDevice
    """
    def __init__(self,
                 stack,
                 **kwa):
        """
        Initialization method for instance

        Inherited Parameters:
            stack is Stack managing this device required
            name is user friendly name of device
            uid  is unique device id
            ha is device udp host address, a duple (host,port)
            kind is type of device

        Parameters:


        Inherited Attributes:
            .stack is Stack managing this device required
            .name is user friendly name of device
            .uid  is unique device id per channel or site
            .ha is device udp host address, a duple (host,port)
            .kind is type of device

        Attributes:

        """
        super(IpLocalDevice, self).__init__(stack=stack, **kwa)


class IpRemoteDevice(IpDevice, RemoteDevice):
    """
    Ip remote device
    """

    def __init__(self,
                 stack,
                 **kwa):
        """
        Initialization method for instance

        Inherited Parameters:
            stack is Stack managing this device required
            name is user friendly name of device
            uid  is unique device id
            ha is device udp host address, a duple (host,port)
            kind is type of device

        Parameters:


        Inherited Attributes:
            .stack is Stack managing this device required
            .name is user friendly name of device
            .uid  is unique device id per channel or site
            .ha is device udp host address, a duple (host,port)
            .kind is type of device

        Attributes:

        """
        super(IpRemoteDevice, self).__init__(stack=stack, **kwa)


class IpSingleRemoteDevice(IpDevice, SingleRemoteDevice):
    """
    Ip Single Remote Device Class when only one remote in stack .remote
    Affects how uid is assigned
    """

    def __init__(self,
                 stack,
                 **kwa):
        """
        Initialization method for instance

        Inherited Parameters:
            stack is Stack managing this device required
            name is user friendly name of device
            uid  is unique device id
            ha is device udp host address, a duple (host,port)
            kind is type of device

        Parameters:


        Inherited Attributes:
            .stack is Stack managing this device required
            .name is user friendly name of device
            .uid  is unique device id per channel or site
            .ha is device udp host address, a duple (host,port)
            .kind is type of device

        Attributes:

        """
        super(IpSingleRemoteDevice, self).__init__(stack=stack, **kwa)



import math
import statistics


def fuzzyAnd(m):
    """
    fuzzy anding
    m = list of membership values to be anded

    returns smallest value in the list
    """
    return min(m)

FuzzyAnd = fuzzyAnd


def fuzzyOr(m):
    """
    fuzzy oring
    m = list of membership values to be ored

    returns largest value in the list
    """
    return max(m)

FuzzyOr = fuzzyOr


def fuzzyNot(x):
    """
    fuzzy not
    x = single membership value to be noted

    returns the inverse membership value
    """
    return 1 - x


def compensatoryAnd(m, g=0.5):
    """
       anding function
       m = list of membership values for x derived from n membership functions
       g = gamma value 0=product 1=algebraic sum

       returns compensatory AND value of x
    """
    g = float(g)

    product1 = 1
    product2 = 1
    for mem in m:
        product1 *= mem
        product2 *= (1 - mem)

    return math.pow(product1, 1 - g) * math.pow((1 - product2), g)

CompensatoryAnd = compensatoryAnd


def gowa(w, wm, l=1.0):
    """
       Generalized Ordered Weighted Averaging Operator
       More info can be found here:
        https://pdfs.semanticscholar.org/2810/c971af0d01d085c799fb2295dc5668d055c8.pdf

       l = -1 = Ordered Weighted Harmonic Averaging Operator
       l = -.000000000001 = Ordered Weighted Geometric Averaging Operator
       l = 1 = Ordered Weighted Arithmetic Averaging Operator
       l = 2 = Ordered Weighted Quadratic Averaging Operator

       w = list of weights
       wm = list of importance weighted membership values
       l = lambda real number specifying type of owa to use

       returns ordered weighted average
    """
    if len(w) != len(wm):
        raise ValueError("Weights and membership value lists must be of equal length.")

    if l == 0:
        raise ZeroDivisionError("Param l cannot be 0.  Use -.000000000001 for owg.")

    wm.sort(reverse=True)

    s = 0
    for i in range(len(w)):
        s += w[i] * math.pow(wm[i], l)

    return math.pow(s, 1/l)

Gowa = gowa


def owa(w, wm):
    """
       Ordered Weighted Arithmetic Averaging Operator
       w = [1,0,0,0] = AND
       w = [0,0,0,1] = OR
       w = [1/n,1/n,1/n,1/n] = Arithmetic Average where n=len(w)

       w = list of weights
       wm = list of importance weighted membership values

       returns ordered arithmetic weighted average
    """

    if len(w) != len(wm):
        raise ValueError("Weights and membership value lists must be of equal length.")

    wm.sort(reverse=True)

    s = 0
    for i in range(len(w)):
        s += w[i] * wm[i]

    return s

Owa = owa


def owg(w, wm):
    """
       Ordered Weighted Geometric Averaging Operator
       More info can be found here:
        ftp://decsai.ugr.es/pub/arai/tech_rep/decision/libroOWG.pdf
       w = [1,0,0,0] = AND
       w = [0,0,0,1] = OR
       w = [1/n,1/n,1/n,1/n] = Geometric Average where n=len(w)

       w = list of weights
       wm = list of importance weighted membership values

       returns ordered geometric weighted average
    """
    if len(w) != len(wm):
        raise ValueError("Weights and membership value lists must be of equal length.")

    wm.sort(reverse=True)

    s = 1
    for i in range(len(w)):
        s *= math.pow(wm[i], w[i])

    return s

Owg = owa


def owh(w, wm):
    """
       Ordered Weighted Harmonic Averaging Operator
       w = [1,0,0,0] = AND
       w = [0,0,0,1] = OR
       w = [1/n,1/n,1/n,1/n] = Harmonic Average where n=len(w)

       w = list of weights
       wm = list of importance weighted membership values

       returns ordered harmonic weighted average
    """
    return gowa(w, wm, -1)

Owh = owh


def owq(w, wm):
    """
       Ordered Weighted Quadratic Averaging Operator
       w = [1,0,0,0] = AND
       w = [0,0,0,1] = OR
       w = [1/n,1/n,1/n,1/n] = Quadratic Average where n=len(w)

       w = list of weights
       wm = list of importance weighted membership values

       returns ordered quadratic weighted average
    """
    return gowa(w, wm, 2)

Owq = owq


def median(wm):
    """
       Median Operator

       wm = list of importance weighted membership values

       returns the middle value in the set
    """
    return statistics.median(wm)

Median = median



"""
eventing.py event and event tagging utility functions

"""
from __future__ import absolute_import, division, print_function

import sys
import os
import datetime
from collections.abc import Set  # both set and frozen set

# Import ioflo libs
from .sixing import *
from .odicting import odict
from .timing import iso8601, tuuid
from .consoling import getConsole

console = getConsole()

def tagify(tail=u'', head=u'', sep=u'.'):
    """
    Returns namespaced event tag string.
    Tag generated by joining with sep the head and tail in that order
    head and tail may be a string or a list, tuple, or Set of strings

    If head is a list, tuple or Set Then
        join with sep all elements of head individually
    Else
        join in whole as string prefix

    If tail is a list, tuple or Set Then
        join with sep all elements of tail individually
    Else
        join in whole as string suffix

    If either head or tail is empty then do not exhibit in tag

    """
    if isinstance(head, (list, tuple, Set)):  # list like so expand
        parts = list(head)
    else:  # string like so put in list
        parts = [head]
    if isinstance(tail, (list, tuple, Set)):  # listlike so extend parts
        parts.extend(tail)
    else:  # string like so append
        parts.append(tail)
    return sep.join([part for part in parts if part])

def eventify(tag, data=None, stamp=None, uid=None):
    """
    Returns new event with tag and  current timestamp and data if any
    Timestamp is iso 8601
        YYYY-MM-DDTHH:MM:SS.mmmmmm which is strftime '%Y-%m-%dT%H:%M:%S.%f'
    Adds uid field if provided
    if not provided data is empty odict
    """
    event = odict([(u'tag', tag),
                   (u'stamp', stamp if stamp is not None else iso8601()),
                   ])
    if uid is not None:
        event['uid'] = uid
    event['data'] = data if data is not None else odict()
    return event

def eventize(tag, stamp=None, uid=None, data=None, route=None):
    """
    Returns odict with fields suitable for event message

    tag is required and must be provided namespaced (dotted) string
    stamp is optional timestamp, if stamp is True then substitude iso8601 time stamp
    uid is optional unique id, if uid is True then substitute time unique identifier
    data is optional data odict, if data is True then substitute empty odict
    route is optional routing odict of form {src: tuple, dst: tuple}
    """
    event = odict(tag=tag)
    if stamp is not None:
        event["stamp"] = iso8601() if stamp is True else stamp
    if uid is not None:
        event['uid'] = tuuid() if uid is True else uid
    if data is  not None:
        event["data"] = odict() if data is True else data
    if route is not None:
        event["route"] = route
    return event



# -*- coding: utf-8 -*-
"""
Unittests for aioing module

"""

import sys
import unittest

import os
import tempfile
import shutil

from ioflo.aid.sixing import *
from ioflo.aid.consoling import getConsole
from ioflo.aio import aioing

console = getConsole()


def setUpModule():
    console.reinit(verbosity=console.Wordage.concise)

def tearDownModule():
    console.reinit(verbosity=console.Wordage.concise)

class BasicTestCase(unittest.TestCase):
    """
    Test Case
    """

    def setUp(self):
        """

        """
        pass


    def tearDown(self):
        """

        """
        pass

    def testNormalizeHost(self):
        """
        Test normalizeHost function
        """
        console.terse("{0}\n".format(self.testNormalizeHost.__doc__))

        host = aioing.normalizeHost("")
        self.assertEqual(host, "0.0.0.0")
        host = aioing.normalizeHost("::")
        self.assertEqual(host, "::")


        host = aioing.normalizeHost("216.58.193.78")
        self.assertEqual(host, "216.58.193.78")
        host = aioing.normalizeHost("2607:f8b0:400a:809::200e")
        self.assertEqual(host, '2607:f8b0:400a:809::200e')


def runOne(test):
    '''
    Unittest Runner
    '''
    test = BasicTestCase(test)
    suite = unittest.TestSuite([test])
    unittest.TextTestRunner(verbosity=2).run(suite)

def runSome():
    """ Unittest runner """
    tests =  []
    names = [
             'testNormalizeHost',
            ]
    tests.extend(map(BasicTestCase, names))
    suite = unittest.TestSuite(tests)
    unittest.TextTestRunner(verbosity=2).run(suite)

def runAll():
    """ Unittest runner """
    suite = unittest.TestSuite()
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(BasicTestCase))
    unittest.TextTestRunner(verbosity=2).run(suite)

if __name__ == '__main__' and __package__ is None:

    #console.reinit(verbosity=console.Wordage.concise)

    #runAll() #run all unittests

    runSome()#only run some

    #runOne('testClientAutoReconnect')


